from pydantic import Field

from solaredge2mqtt.core.models import BaseInputField, BaseInputFieldEnumModel


class ModbusActivePowerLimitInput(BaseInputField):
    limit: int = Field(min=0, max=100)


class ModbusPowerControlInput(BaseInputFieldEnumModel):
    ACTIVE_POWER_LIMIT = "active_power_limit", ModbusActivePowerLimitInput



from __future__ import annotations

from collections.abc import MutableMapping
from datetime import datetime, timezone
from enum import Enum

import jsonref
from pydantic import BaseModel, model_serializer, model_validator

from solaredge2mqtt import __version__


class EnumModel(Enum):
    def __new__(cls, *args):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __str__(self) -> str:
        return str(self._value_)

    def __repr__(self) -> str:
        return str(self._value_)

    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__) and self.value == other.value

    def __hash__(self):
        return hash(self.value)

    @classmethod
    def from_string(cls, value: str) -> EnumModel:
        for item in cls.__members__.values():
            if item.value == value:
                return item
        raise ValueError(f"No enum value {value} found.")

    @model_serializer
    def serialize(self) -> any:
        return self.value


class BaseInputField(BaseModel):
    model_config = {
        "extra": "forbid"
    }


class BaseInputFieldEnumModel(EnumModel):
    def __init__(self, key: str, input_model: type[BaseInputField]):
        self._key: str = key
        self._input_model: type[BaseInputField] = input_model

    @property
    def key(self) -> str:
        return self._key

    @property
    def input_model(self) -> type[BaseInputField]:
        return self._input_model


class BaseField(EnumModel):
    def __init__(self, key: str, input_field: BaseInputFieldEnumModel | None = None):
        self._key: str = key
        self._input_field: BaseInputFieldEnumModel | None = input_field

    def field(self, title: str | None,
              json_schema_extra: dict[str, any] | None = None,
              input_field: BaseInputFieldEnumModel | None = None) -> dict[str, any]:

        if json_schema_extra is None:
            json_schema_extra = {}

        input_field = input_field or self._input_field

        if input_field:
            json_schema_extra = {
                "input_field": input_field.key,
                **json_schema_extra,
            }
        else:
            json_schema_extra = {
                "input_field": None,
                **json_schema_extra,
            }

        return {"title": title, "json_schema_extra": json_schema_extra}


class Solaredge2MQTTBaseModel(BaseModel):
    timestamp: datetime | None = None

    @model_validator(mode="before")
    def _set_always(cls, data: dict[str, any]) -> dict[str, any]:
        data = dict(data or {})
        if "timestamp" not in data or data["timestamp"] is None:
            data["timestamp"] = datetime.now(tz=timezone.utc)
        return data

    def model_dump_influxdb(self, exclude: list[str] | None = None) -> dict[str, any]:
        ignore_keys = {"timestamp"}
        return self._flatten_dict(
            self.model_dump(exclude=exclude, exclude_none=True),
            ignore_keys=ignore_keys
        )

    def _flatten_dict(
        self,
        d: MutableMapping,
        ignore_keys: set[str] = set(),
        join_chr: str = "_",
        parent_key: str = "",
    ) -> MutableMapping:
        items = []
        for k, v in d.items():
            if k in ignore_keys:
                continue
            new_key = parent_key + join_chr + k if parent_key else k
            if isinstance(v, MutableMapping):
                items.extend(
                    self._flatten_dict(
                        v, ignore_keys, join_chr, new_key
                    ).items()
                )
            else:
                if isinstance(v, int):
                    v = float(v)
                elif isinstance(v, datetime):
                    v = v.isoformat()

                items.append((new_key, v))
        return dict(items)

    def _default_homeassistant_device_info(self, name: str) -> dict[str, any]:
        return {
            "name": f"SolarEdge2MQTT {name}",
            "manufacturer": "DerOetzi",
            "model": "SolarEdge2MQTT",
            "sw_version": __version__,
        }

    @classmethod
    def parse_schema(cls, property_parser: callable | None = None) -> list[dict]:
        return cls._walk_schema(
            jsonref.replace_refs(
                cls.model_json_schema(mode="serialization"),
                merge_props=True
            )[
                "properties"
            ], property_parser or cls.property_parser

        )

    @classmethod
    def _walk_schema(
        cls,
        properties: dict[str, dict],
        property_parser: callable,
        parent_name: str | None = None,
        parent_path: list[str] = []
    ) -> list[dict]:
        items: list[dict] = []
        for key, prop in properties.items():
            new_path = [*parent_path, key]
            new_name = (
                parent_name + " " +
                prop["title"] if parent_name else prop["title"]
            )
            if "properties" in prop:
                items.extend(
                    cls._walk_schema(
                        prop["properties"], property_parser, new_name, new_path
                    )
                )
            elif "allOf" in prop:
                items.extend(
                    cls._walk_schema(
                        prop["allOf"][0]["properties"],
                        property_parser,
                        new_name,
                        new_path
                    )
                )
            elif "anyOf" in prop and "properties" in prop["anyOf"][0]:
                items.extend(
                    cls._walk_schema(
                        prop["anyOf"][0]["properties"],
                        property_parser,
                        new_name,
                        new_path
                    )
                )
            else:
                item = property_parser(prop, new_name, new_path)
                if item:
                    items.append(item)

        return items

    @staticmethod
    def property_parser(
        prop: dict[str, any],
        name: str,
        path: list[str],
    ) -> dict | None:
        if "input_field" in prop:
            return {
                "name": name,
                "path": path,
                "input_field": prop["input_field"],
            }



from __future__ import annotations

import importlib.util

_FORECAST_DEPS = ("numpy", "pandas", "scipy", "sklearn")


def _deps_available() -> bool:
    return all(importlib.util.find_spec(pkg) is not None for pkg in _FORECAST_DEPS)


FORECAST_AVAILABLE: bool = _deps_available()


if FORECAST_AVAILABLE:
    from .service import ForecastService
    __all__ = ["ForecastService"]
else:
    class ForecastService:
        def __init__(self, *args, **kwargs):
            raise ImportError(
                (
                    "ForecastService requires to install extra dependencies. "
                    "Run pip install solaredge2mqtt[forecast]"
                )
            )



from __future__ import annotations

from datetime import datetime

from pydantic import Field, field_serializer, model_serializer

from solaredge2mqtt.core.models import Solaredge2MQTTBaseModel


class OpenWeatherMapOneCallBase(Solaredge2MQTTBaseModel):
    lat: float
    lon: float
    timezone: str
    timezone_offset: int


class OpenWeatherMapOneCall(OpenWeatherMapOneCallBase):
    current: OpenWeatherMapCurrentData
    hourly: list[OpenWeatherMapForecastData]


class OpenWeatherMapRain(Solaredge2MQTTBaseModel):
    one_hour: float = Field(0.0, alias="1h")

    @model_serializer
    def ser_model(self) -> float:
        return self.one_hour


class OpenWeatherMapSnow(OpenWeatherMapRain):
    pass


class OpenWeatherMapBaseData(Solaredge2MQTTBaseModel):
    dt: datetime
    temp: float | None = Field(None)
    feels_like: float | None = Field(None)
    pressure: int | None = Field(None)
    humidity: int | None = Field(None)
    dew_point: float | None = Field(None)
    uvi: float | None = Field(None)
    clouds: int | None = Field(None)
    visibility: int | None = Field(None)
    wind_speed: float | None = Field(None)
    wind_deg: int | None = Field(None)
    wind_gust: float | None = Field(None)
    weather: list[OpenWeatherMapCondition]
    rain: OpenWeatherMapRain = Field(OpenWeatherMapRain())
    snow: OpenWeatherMapSnow = Field(OpenWeatherMapSnow())

    @property
    def localtime(self) -> datetime:
        return self.dt.astimezone()

    @property
    def year(self) -> int:
        return self.localtime.year

    @property
    def month(self) -> int:
        return self.localtime.month

    @property
    def day(self) -> int:
        return self.localtime.day

    @property
    def hour(self) -> int:
        return self.localtime.hour

    @field_serializer("dt")
    def serialize_dt(self, dt: datetime, _info) -> str:
        return dt.astimezone().isoformat()

    @field_serializer("weather")
    def serialize_weather(
        self, weather: list[OpenWeatherMapCondition], _info
    ) -> OpenWeatherMapCondition:
        return weather[0]

    def model_dump_estimation_data(self) -> dict[str, str | float | int | None]:
        model_dict = self.model_dump(exclude=["weather", "dt"], exclude_none=True)
        model_dict["weather_id"] = self.weather[0].id
        model_dict["weather_main"] = self.weather[0].main
        return model_dict


class OpenWeatherMapCurrentData(OpenWeatherMapBaseData):
    sunrise: datetime = Field(exclude=True)
    sunset: datetime = Field(exclude=True)


class OpenWeatherMapForecastData(OpenWeatherMapBaseData):
    pop: float


class OpenWeatherMapCondition(Solaredge2MQTTBaseModel):
    id: int
    main: str
    description: str
    icon: str = Field(exclude=True)



from __future__ import annotations

import asyncio
from typing import Awaitable, Callable

from aiomqtt import MqttError

from solaredge2mqtt.core.events.events import BaseEvent
from solaredge2mqtt.core.exceptions import InvalidDataException
from solaredge2mqtt.core.logging import logger

Listener = Callable[[BaseEvent], Awaitable[None]]


class EventBus:
    def __init__(self) -> None:
        self._listeners: dict[str, list[Listener]] = {}
        self._subscribed_events: dict[str, type[BaseEvent]] = {}
        self._tasks: set[asyncio.Task] = set()

    def subscribe(
        self,
        event: type[BaseEvent] | list[type[BaseEvent]],
        listener: Listener,
    ) -> None:
        if isinstance(event, list):
            for _event in event:
                self.subscribe(_event, listener)
            return
        event_key = event.event_key()
        logger.info(f"Event subscribed: {event_key}")
        self._listeners.setdefault(event_key, []).append(listener)
        self._subscribed_events[event_key] = event

    @property
    def subscribed_events(self) -> list[type[BaseEvent]]:
        return list(self._subscribed_events.values())

    def unsubscribe(self, event: type[BaseEvent], listener: Listener) -> None:
        event_key = event.event_key()
        if event_key in self._listeners:
            try:
                self._listeners[event_key].remove(listener)
            except ValueError:
                pass
            if not self._listeners[event_key]:
                self._listeners.pop(event_key, None)
                self._subscribed_events.pop(event_key, None)

    def unsubscribe_all(self, event: type[BaseEvent]) -> None:
        event_key = event.event_key()
        if event_key in self._listeners:
            self._listeners.pop(event_key, None)
            self._subscribed_events.pop(event_key, None)

    async def emit(self, event: BaseEvent) -> None:
        event_key = event.event_key()
        logger.trace(f"Event emitted: {event_key}")
        listeners = self._listeners.get(event_key)
        if not listeners:
            return
        if event.AWAIT:
            await self._notify_listeners(event, listeners)
        else:
            task = asyncio.create_task(
                self._notify_listeners(event, listeners))
            self._tasks.add(task)
            task.add_done_callback(lambda t: self._tasks.discard(t))

    async def _notify_listeners(
        self, event: BaseEvent, listeners: list[Listener]
    ) -> None:
        results = await asyncio.gather(
            *(self._notify_listener(listener, event)
              for listener in listeners),
            return_exceptions=True,
        )
        for r in results:
            if isinstance(r, (asyncio.CancelledError, MqttError)):
                raise r
            elif isinstance(r, Exception):
                logger.error("Unhandled listener error: {exc}", exc=repr(r))

    async def _notify_listener(self, listener: Listener, event: BaseEvent) -> None:
        try:
            await listener(event)
        except InvalidDataException as error:
            logger.warning("{message}, skipping this loop",
                           message=error.message)

    async def cancel_tasks(self) -> None:
        tasks = list(self._tasks)
        for t in tasks:
            t.cancel()

        await asyncio.gather(*tasks, return_exceptions=True)

        self._tasks.clear()

        logger.info("Running tasks cancelled: {count}", count=len(tasks))



from __future__ import annotations

from functools import lru_cache

from solaredge2mqtt.core.settings.models import ServiceSettings


@lru_cache()
def service_settings() -> ServiceSettings:
    return ServiceSettings()



from datetime import datetime

from pydantic import computed_field

from solaredge2mqtt.core.models import EnumModel
from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantSensorType as HASensor,
)
from solaredge2mqtt.services.models import Component


class ForecasterType(EnumModel):
    ENERGY = "energy"
    POWER = "power"

    def __init__(self, target_column: str) -> None:
        self._target_column: str = target_column

    @property
    def target_column(self) -> str:
        return self._target_column

    def prepare_value(self, value: float | int) -> float | int:
        if value <= 0:
            prepared = 0
        elif self.target_column == "energy":
            prepared = round(value / 1000, 3)
        else:
            prepared = int(round(value))

        return prepared


class Forecast(Component):
    COMPONENT = "forecast"

    power_period: dict[datetime, int]
    energy_period: dict[datetime, int]

    @computed_field(**HASensor.ENERGY_WH.field("Energy production today"))
    @property
    def energy_today(self) -> int:
        return sum(self._energy_today)

    @computed_field(**HASensor.ENERGY_WH.field("Energy production remaining today"))
    @property
    def energy_today_remaining(self) -> int:
        return sum(self._energy_today[self._current_hour():])

    @computed_field(**HASensor.ENERGY_WH.field("Energy production current hour"))
    def energy_current_hour(self) -> int:
        return self._energy_today[self._current_hour()]

    @computed_field(**HASensor.ENERGY_WH.field("Energy production next hour"))
    def energy_next_hour(self) -> int:
        if self._current_hour() == 23:
            energy_next_hour = self._energy_tomorrow[0]
        else:
            energy_next_hour = self._energy_today[self._current_hour() + 1]

        return energy_next_hour

    @computed_field(**HASensor.ENERGY_WH.field("Energy production tomorrow"))
    @property
    def energy_tomorrow(self) -> int:
        return sum(self._energy_tomorrow)

    @property
    def _energy_today(self) -> list[int]:
        return [*self.energy_period.values()][:24]

    @property
    def _energy_tomorrow(self) -> list[int]:
        return [*self.energy_period.values()][24:]

    @staticmethod
    def _current_hour() -> int:
        return datetime.now().hour

    @classmethod
    # pylint: disable=arguments-differ
    def model_json_schema(cls, mode: str = "serialization") -> dict[str, any]:
        schema = super().model_json_schema(mode=mode)
        schema["properties"].pop("power_period", None)
        schema["properties"].pop("energy_period", None)
        return schema

    def homeassistant_device_info(self) -> dict[str, any]:
        return self._default_homeassistant_device_info("Forecast")



from pydantic import BaseModel

from solaredge2mqtt.core.events.events import BaseEvent
from solaredge2mqtt.core.models import BaseInputField


class MQTTPublishEvent(BaseEvent):
    AWAIT = True

    def __init__(
        self,
        topic: str,
        payload: str | int | float | BaseModel,
        retain: bool,
        qos: int = 0,
        topic_prefix: str | None = None,
        exclude_none: bool = False,
    ):
        self._topic: str = topic
        self._payload: str | int | float | BaseModel = payload
        self._retain: bool = retain
        self._qos: int = qos
        self._topic_prefix: str | None = topic_prefix
        self._exclude_none: bool = exclude_none

    @property
    def topic(self) -> str:
        return self._topic

    @property
    def payload(self) -> str | int | float | BaseModel:
        return self._payload

    @property
    def retain(self) -> bool:
        return self._retain

    @property
    def qos(self) -> int:
        return self._qos

    @property
    def topic_prefix(self) -> str | None:
        return self._topic_prefix

    @property
    def exclude_none(self) -> bool:
        return self._exclude_none


class MQTTReceivedEvent(BaseEvent):
    def __init__(self, topic: str, input: BaseInputField):
        self._topic: str = topic
        self._input: BaseInputField = input

    @property
    def topic(self) -> str:
        return self._topic

    @property
    def input(self) -> BaseInputField:
        return self._input


class MQTTSubscribeEvent(BaseEvent):
    AWAIT = True

    def __init__(self, topic: str, model: type[BaseInputField]):
        self._topic: str = topic
        self._model: type[BaseInputField] = model

    @property
    def topic(self) -> str:
        return self._topic

    @property
    def model(self) -> type[BaseInputField]:
        return self._model



from pydantic import Field

from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantSensorType as HASensor,
)
from solaredge2mqtt.services.modbus.models.base import ModbusComponent, ModbusDeviceInfo
from solaredge2mqtt.services.modbus.models.values import (
    ModbusACCurrent,
    ModbusACPower,
    ModbusACVoltage,
    ModbusEnergy,
)


class ModbusMeter(ModbusComponent):
    COMPONENT = "meter"

    current: ModbusACCurrent = Field(title="Current")
    voltage: ModbusACVoltage = Field(title="Voltage")
    power: ModbusACPower = Field(title="Power")
    energy: ModbusEnergy = Field(title="Energy")
    frequency: float = Field(**HASensor.FREQUENCY_HZ.field("Grid frequency"))

    def __init__(self, info: ModbusDeviceInfo, data: dict[str, str | int]):
        current = ModbusACCurrent(data)
        voltage = ModbusACVoltage(data)
        power = ModbusACPower(data, "power")
        energy = ModbusEnergy(data)
        frequency = self.scale_value(data, "frequency")

        super().__init__(
            info=info,
            current=current,
            voltage=voltage,
            power=power,
            energy=energy,
            frequency=frequency,
        )

    def homeassistant_device_info_with_name(self, name: str) -> dict[str, any]:
        return self.info.homeassistant_device_info(name)



from pathlib import Path

import platformdirs
from pydantic import BaseModel, DirectoryPath, Field, field_validator


class ForecastSettings(BaseModel):
    enable: bool = Field(False)
    hyperparametertuning: bool = Field(False)
    cachingdir: DirectoryPath | None = Field(
        default=str(Path(platformdirs.user_cache_dir("se2mqtt_forecast")))
    )
    retain: bool = Field(False)

    @property
    def is_configured(self) -> bool:
        return self.enable

    @property
    def is_caching_enabled(self) -> bool:
        return self.cachingdir is not None

    @field_validator("cachingdir")
    def ensure_secure_cache(cls, v: str | None, values: dict) -> str | None:
        if not values.get("enable", False) or v is None:
            return None

        path = Path(v).resolve()

        path.mkdir(parents=True, exist_ok=True, mode=0o700)

        if path.stat().st_mode & 0o077:
            raise ValueError(f"Insecure cache directory permissions: {path}")

        return str(path)



from solaredge2mqtt.core.events.events import BaseEvent


class IntervalBaseTriggerEvent(BaseEvent):
    pass


class Interval1MinTriggerEvent(BaseEvent):
    pass


class Interval5MinTriggerEvent(BaseEvent):
    pass


class Interval10MinTriggerEvent(BaseEvent):
    pass


class Interval15MinTriggerEvent(BaseEvent):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTReceivedEvent
from solaredge2mqtt.services.modbus.events import ModbusWriteEvent
from solaredge2mqtt.services.modbus.models.base import ModbusUnitRole
from solaredge2mqtt.services.modbus.models.inputs import ModbusPowerControlInput
from solaredge2mqtt.services.modbus.models.inverter import ModbusInverter
from solaredge2mqtt.services.modbus.settings import AdvancedControlsSettings
from solaredge2mqtt.services.modbus.sunspec.inverter import SunSpecPowerControlRegister

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings.models import ServiceSettings


class ModbusAdvancedControl:
    def __init__(self, service_settings: ServiceSettings, event_bus: EventBus):
        self.settings = service_settings.modbus

        inverter_topic = (
            ModbusInverter.generate_topic_prefix(str(ModbusUnitRole.LEADER))
            if self.settings.has_followers
            else ModbusInverter.generate_topic_prefix())

        self.topic_prefix = f"{service_settings.mqtt.topic_prefix}/{inverter_topic}"
        self.event_bus = event_bus

        self._subscribe_events()

    def _subscribe_events(self) -> None:
        if self.settings.advanced_power_controls_enabled:
            self.event_bus.subscribe(
                MQTTReceivedEvent, self.handle_mqtt_received_event
            )

    async def async_init(self) -> None:
        await self.handle_advanced_power_control_settings()

    async def handle_advanced_power_control_settings(self):
        if self.settings.advanced_power_controls == AdvancedControlsSettings.ENABLED:
            await self.enable_advanced_power_control()
            await self.subscribe_topics()
        elif self.settings.advanced_power_controls == AdvancedControlsSettings.DISABLE:
            await self.disable_advanced_control_settings()

            logger.warning(
                "Change setting to disabled and restart the service.")
        else:
            logger.info("Advanced power control is disabled in settings")

    def enable_advanced_power_control(self):
        logger.debug("Enabling advanced power control")
        # await self.event_bus.emit(ModbusWriteEvent(
        #    SunSpecPowerControlRegister.ADVANCED_POWER_CONTROL_ENABLE, True
        # ))
        logger.info("Advanced power control enabled.")

    async def disable_advanced_control_settings(self):
        logger.debug("Disabling advanced power control")
        await self.event_bus.emit(ModbusWriteEvent(
            SunSpecPowerControlRegister.REACTIVE_POWER_CONFIG, 0
        ))

        await self.event_bus.emit(ModbusWriteEvent(
            SunSpecPowerControlRegister.ADVANCED_POWER_CONTROL_ENABLE, False
        ))

        await self.event_bus.emit(ModbusWriteEvent(
            SunSpecPowerControlRegister.COMMIT_POWER_CONTROL_SETTINGS, 1
        ))

        logger.info("Advanced power control disabled.")

    def subscribe_topics(self) -> None:
        for field in ModbusInverter.parse_schema(self.property_parser):
            topic = f"{self.topic_prefix}/{field['topic']}"
            logger.info(f"Subscribing to topic: {topic}")
            # await self.event_bus.emit(MQTTSubscribeEvent(topic))

    @staticmethod
    def property_parser(
        prop: dict[str, any],
        name: str,
        path: list[str]
    ) -> dict[str, str] | None:
        field = None

        if prop.get("input_field", False):
            input_field = prop.get("input_field")
            input_field = ModbusPowerControlInput.from_string(input_field)
            field = {
                "name": name,
                "topic": "/".join(path),
                "input_field": input_field,
            }

        return field

    def handle_mqtt_received_event(self, event: MQTTReceivedEvent) -> None:
        logger.info(event.input)



import asyncio
from datetime import timezone

from pandas import read_csv, to_datetime
from tzlocal import get_localzone_name

from solaredge2mqtt.core.influxdb import InfluxDB, Point
from solaredge2mqtt.core.settings import service_settings
from solaredge2mqtt.services.powerflow.models import (
    BatteryPowerflow,
    ConsumerPowerflow,
    GridPowerflow,
    InverterPowerflow,
    Powerflow,
)

settings = service_settings()
STRINGS = 2

LOCAL_TZ = get_localzone_name()


async def main():
    data = read_and_prepare_data()

    points = []

    columns = data.columns.to_list()

    for _, row in data.iterrows():

        grid, battery, pv_production, inverter, consumer, powerflow = create_powerflow(
            columns, row
        )

        print(
            f"Time: {row['time']}, PV: {pv_production} W, "
            f"Inverter: {inverter.power} W, "
            f"House: {consumer.house} W, Grid: {grid.power} W, "
            f"Battery: {battery.power} W, Wallbox: {consumer.evcharger} W"
        )

        prepare_points(points, columns, row, powerflow)

    influxdb = InfluxDB(settings.influxdb, settings.prices)
    await influxdb.write_points_async(points)


def read_and_prepare_data():
    data = read_csv("data.csv")
    data["time"] = to_datetime(data["time"], format="%d.%m.%Y %H:%M").dt.tz_localize(
        LOCAL_TZ
    )

    for i in range(1, STRINGS + 1):
        data[f"string{i}"] = data[f"string{i}"].fillna(0)

    data["evcharger"] = data["evcharger"].fillna(0)

    data.info()
    return data


def create_powerflow(columns, row):
    grid = GridPowerflow(
        power=round(row["grid_delivery"] - row["grid_consumption"])
    )

    if "battery_charge" in columns and "battery_discharge" in columns:
        battery = BatteryPowerflow(
            power=round(row["battery_charge"] - row["battery_discharge"])
        )
    else:
        battery = BatteryPowerflow(power=0)

    pv_production = 0.0
    for i in range(1, STRINGS + 1):
        pv_production += row[f"string{i}"]

    dc_power = pv_production - battery.power

    inverter_power = round(
        row["inverter_production"] - row["inverter_consumption"])

    inverter = InverterPowerflow(
        power=inverter_power,
        dc_power=round(dc_power),
        battery_discharge=battery.discharge,
    )

    if "evcharger" in columns:
        evcharger = round(row["evcharger"])
    else:
        evcharger = 0

    consumer = ConsumerPowerflow(
        inverter=inverter, grid=grid, evcharger=evcharger)

    pv_production = round(pv_production)

    powerflow = Powerflow(
        pv_production=pv_production,
        inverter=inverter,
        grid=grid,
        battery=battery,
        consumer=consumer,
    )

    return grid, battery, pv_production, inverter, consumer, powerflow


def prepare_points(points, columns, row, powerflow):
    point = powerflow.prepare_point("powerflow_import")
    point.tag("agg_type", "mean")
    point.time(row["time"].astimezone(timezone.utc))
    points.append(point)

    point = powerflow.prepare_point_energy("energy_import", settings.prices)
    point.time(row["time"].astimezone(timezone.utc))
    points.append(point)

    if "battery_soc" in columns:
        battery_soc = row["battery_soc"]
        if battery_soc < 100:
            battery_soc += 0.01

        point = Point("battery_import")
        point.field("state_of_charge", round(battery_soc, 2))
        point.time(row["time"].astimezone(timezone.utc))
        point.tag("agg_type", "mean")
        points.append(point)


if __name__ == "__main__":
    asyncio.run(main())



from pydantic import BaseModel, Field


class HomeAssistantSettings(BaseModel):
    enable: bool = Field(False)
    topic_prefix: str = Field("homeassistant")
    retain: bool = Field(False)

    @property
    def is_configured(self) -> bool:
        return self.enable




from influxdb_client import Point
from pydantic import Field

from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantSensorType as HASensor,
)
from solaredge2mqtt.services.modbus.models.base import ModbusComponent, ModbusDeviceInfo
from solaredge2mqtt.services.modbus.sunspec.values import BATTERY_STATUS_MAP


class ModbusBattery(ModbusComponent):
    COMPONENT = "battery"

    status: int = Field(**HASensor.STATUS.field("Status"))
    status_text: str = Field(**HASensor.STATUS.field("Status text"))
    current: float = Field(**HASensor.CURRENT_A.field("current"))
    voltage: float = Field(**HASensor.VOLTAGE_V.field("voltage"))
    power: float = Field(**HASensor.POWER_W.field("power"))
    state_of_charge: float = Field(
        **HASensor.BATTERY.field("state of charge"))
    state_of_health: float = Field(
        **HASensor.BATTERY.field("state of health"))

    def __init__(self, info: ModbusDeviceInfo, data: dict[str, str | int]) -> None:
        status = data["status"]
        if status in BATTERY_STATUS_MAP:
            status_text = BATTERY_STATUS_MAP[data["status"]]
        else:
            status_text = "Unknown"

        current = round(data["instantaneous_current"], 2)
        voltage = round(data["instantaneous_voltage"], 2)
        power = round(data["instantaneous_power"], 2)
        state_of_charge = round(data["soe"], 2)
        state_of_health = round(data["soh"], 2)

        super().__init__(
            info=info,
            status=status,
            status_text=status_text,
            current=current,
            voltage=voltage,
            power=power,
            state_of_charge=state_of_charge,
            state_of_health=state_of_health,
        )

    @property
    def is_valid(self) -> bool:
        valid = False

        if self.state_of_charge < 0:
            logger.warning("Battery state of charge is negative")
        elif self.state_of_health < 0:
            logger.warning("Battery state of health is negative")
        elif self.current < -1000000:
            logger.warning("Battery current is a huge negative")
        else:
            valid = True

        return valid

    def prepare_point(self, measurement: str = "battery_raw") -> Point:
        point = Point(measurement)
        point.field("current", self.current)
        point.field("voltage", self.voltage)
        point.field("state_of_charge", self.state_of_charge)
        point.field("state_of_health", self.state_of_health)

        if self.has_unit:
            point.tag("unit", self.info.unit.key)

        return point

    def homeassistant_device_info_with_name(self, name: str) -> dict[str, any]:
        return self.info.homeassistant_device_info(name)



class ConfigurationException(Exception):
    def __init__(self, component: str, message: str, *args: any) -> None:
        self.component = component
        self.message = message
        super().__init__(*args)


class InvalidDataException(Exception):
    def __init__(self, message: str, *args: any) -> None:
        self.message = message

        super().__init__(*args)



from __future__ import annotations

from pydantic import Field

from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantBinarySensorType as HABinarySensor,
)
from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantNumberType as HANumber,
)
from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantSensorType as HASensor,
)
from solaredge2mqtt.services.modbus.models.base import ModbusComponent, ModbusDeviceInfo
from solaredge2mqtt.services.modbus.models.inputs import ModbusPowerControlInput
from solaredge2mqtt.services.modbus.models.values import ModbusAC, ModbusDC
from solaredge2mqtt.services.modbus.sunspec.values import INVERTER_STATUS_MAP
from solaredge2mqtt.services.models import ComponentValueGroup


class ModbusInverter(ModbusComponent):
    COMPONENT = "inverter"

    ac: ModbusAC = Field(title="AC")
    dc: ModbusDC = Field(title="DC")
    energytotal: float = Field(**HASensor.ENERGY_WH.field("Energy total"))
    temperature: float = Field(**HASensor.TEMP_C.field("Temperature"))
    status_text: str = Field(**HASensor.STATUS.field("Status text"))
    status: int = Field(**HASensor.STATUS.field("status"))
    grid_status: bool | None = Field(
        None, **HABinarySensor.GRID_STATUS.field("Grid status")
    )
    advanced_power_controls: ModbusPowerControl | None = Field(
        None, title="Advanced Power Controls")

    def __init__(self, info: ModbusDeviceInfo, data: dict[str, str | int]):
        ac = ModbusAC(data)
        dc = ModbusDC(data)
        energytotal = self.scale_value(data, "energy_total")

        status = data["status"]
        if status in INVERTER_STATUS_MAP:
            status_text = INVERTER_STATUS_MAP[status]
        else:
            status_text = "Unknown"

        temperature = self.scale_value(data, "temperature")

        grid_status = None
        if "grid_status" in data:
            grid_status = not data["grid_status"]

        advanced_power_controls = None
        if (
            "advanced_power_control_enable" in data
            and data["advanced_power_control_enable"]
        ):
            advanced_power_controls = ModbusPowerControl(data)

        super().__init__(
            info=info,
            ac=ac,
            dc=dc,
            energytotal=energytotal,
            temperature=temperature,
            status=status,
            status_text=status_text,
            grid_status=grid_status,
            advanced_power_controls=advanced_power_controls,
        )

    def homeassistant_device_info(self) -> dict[str, any]:
        return self.info.homeassistant_device_info("Inverter")


class ModbusPowerControl(ComponentValueGroup):
    advanced_power_control: bool = Field(
        **HABinarySensor.ENABLED.field("Control enabled"))
    active_power_limit: int = Field(
        **HANumber.ACTIVE_POWER_LIMIT.field(
            ModbusPowerControlInput.ACTIVE_POWER_LIMIT,
            "Active PowerLimit"
        ))

    def __init__(self, data: dict[str, str | int | bool]):
        super().__init__(
            advanced_power_control=data["advanced_power_control_enable"],
            active_power_limit=data["active_power_limit"],
        )



from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel, Field, computed_field


class LogicalInfo(BaseModel):
    identifier: str
    serialnumber: str | None
    name: str
    type: str

    @staticmethod
    def map(data: dict[str, str | int]) -> dict[str, str]:
        return {
            "identifier": str(data["id"]),
            "serialnumber": data["serialNumber"],
            "name": data["name"],
            "type": data["type"],
        }


class LogicalInverter(BaseModel):
    info: LogicalInfo
    energy: float | None
    strings: list[LogicalString] = []


class LogicalString(BaseModel):
    info: LogicalInfo
    energy: float | None
    modules: list[LogicalModule] = []


class LogicalModule(BaseModel):
    info: LogicalInfo
    energy: float | None = Field(None)
    power: dict[datetime, float] | None = Field(None)

    @computed_field
    @property
    def power_today(self) -> dict[str, float] | None:
        power_today = None

        if self.power:
            power_today = {
                k.strftime("%H:%M"): v
                for k, v in self.power.items()
            }

        return power_today



"""
    This module, service.py, is part of the SolarEdge2MQTT service, which reads data 
    from a SolarEdge inverter and publishes it to an MQTT broker. It uses the asyncio 
    library for asynchronous I/O and the aiomqtt library for MQTT communication. 
    The module also includes a run function to initialize and start the service.
"""

import asyncio
import platform
import signal
from time import time
from typing import Callable

from aiomqtt import MqttError
from tzlocal import get_localzone_name

from solaredge2mqtt import __version__
from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import ConfigurationException
from solaredge2mqtt.core.influxdb import InfluxDBAsync
from solaredge2mqtt.core.logging import initialize_logging, logger
from solaredge2mqtt.core.mqtt import MQTTClient
from solaredge2mqtt.core.settings import service_settings
from solaredge2mqtt.core.timer import Timer
from solaredge2mqtt.services.energy import EnergyService
from solaredge2mqtt.services.forecast import FORECAST_AVAILABLE
from solaredge2mqtt.services.homeassistant import HomeAssistantDiscovery
from solaredge2mqtt.services.monitoring import MonitoringSite
from solaredge2mqtt.services.powerflow import PowerflowService
from solaredge2mqtt.services.weather import WeatherClient

if FORECAST_AVAILABLE:
    from solaredge2mqtt.services.forecast import ForecastService

LOCAL_TZ = get_localzone_name()


def run():
    try:
        service = Service()
        loop = asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGINT, service.cancel)
        loop.add_signal_handler(signal.SIGTERM, service.cancel)
        loop.run_until_complete(service.main_loop())
    except ConfigurationException:
        logger.error("Configuration error")
    except asyncio.CancelledError:
        logger.debug("Service cancelled")
    finally:
        loop.close()


class Service:
    def __init__(self):
        self.settings = service_settings()
        initialize_logging(self.settings.logging_level)
        logger.debug(self.settings)

        self.event_bus = EventBus()
        self.timer = Timer(self.event_bus, self.settings.interval)

        self.mqtt: MQTTClient | None = None

        self.cancel_request = asyncio.Event()
        self.loops: set[asyncio.Task] = set()

        self.influxdb: InfluxDBAsync | None = (
            InfluxDBAsync(self.settings.influxdb,
                          self.settings.prices, self.event_bus)
            if self.settings.is_influxdb_configured
            else None
        )

        self.energy: EnergyService | None = (
            EnergyService(self.settings.energy, self.event_bus, self.influxdb)
            if self.settings.is_influxdb_configured
            else None
        )

        self.powerflow = PowerflowService(
            self.settings, self.event_bus, self.influxdb)

        self.monitoring: MonitoringSite | None = (
            MonitoringSite(self.settings.monitoring,
                           self.event_bus, self.influxdb)
            if self.settings.is_monitoring_configured
            else None
        )

        self.weather: WeatherClient | None = (
            WeatherClient(self.settings, self.event_bus)
            if self.settings.is_weather_configured
            else None
        )

        if FORECAST_AVAILABLE:
            self.forecast: ForecastService | None = (
                ForecastService(
                    self.settings.forecast,
                    self.settings.location,
                    self.event_bus,
                    self.influxdb,
                )
                if self.settings.is_forecast_configured
                else None
            )
        elif self.settings.is_forecast_configured:
            logger.warning(
                "Forecast service not available, please refer to README")

        self.homeassistant: HomeAssistantDiscovery | None = (
            HomeAssistantDiscovery(self.settings, self.event_bus)
            if self.settings.is_homeassistant_configured
            else None
        )

    def cancel(self):
        logger.info("Stopping SolarEdge2MQTT service...")

        self.cancel_request.set()

        for task in list(self.loops):
            task.cancel()

    async def main_loop(self):
        logger.info("Starting SolarEdge2MQTT service...")
        logger.info("Version: {version}", version=__version__)
        logger.info(
            f"Operationg system: {platform.platform()} "
            f"({platform.system()}/{platform.machine()})"
        )
        logger.debug(self.settings)
        logger.info("Timezone: {timezone}", timezone=LOCAL_TZ)

        if self.settings.is_influxdb_configured:
            self.influxdb.init()

        while not self.cancel_request.is_set():
            try:
                self.mqtt = MQTTClient(self.settings.mqtt, self.event_bus)

                async with self.mqtt:
                    await self.mqtt.publish_status_online()

                    if self.settings.is_homeassistant_configured:
                        await self.homeassistant.async_init()

                    await self.powerflow.async_init()

                    self._start_mqtt_listener()
                    self.schedule_loop(1, self.timer.loop)

                    await asyncio.gather(*self.loops)
            except MqttError:
                logger.error("MQTT error, reconnecting in 5 seconds...")
            except asyncio.CancelledError:
                logger.debug("Loops cancelled")
                raise
            finally:
                await self.finalize()

            if self.cancel_request.is_set():
                break

            await asyncio.sleep(5)

    async def finalize(self):
        try:
            await self.mqtt.publish_status_offline()
        except MqttError:
            pass

        await self.event_bus.cancel_tasks()

        await self.close()

    def _start_mqtt_listener(self):
        task = asyncio.create_task(self.mqtt.listen())
        self.loops.add(task)
        task.add_done_callback(self.loops.remove)

        task = asyncio.create_task(self.mqtt.process_queue())
        self.loops.add(task)
        task.add_done_callback(self.loops.remove)

    def schedule_loop(
        self,
        interval_in_seconds: int,
        handles: Callable | list[Callable],
        delay_start: int = 0,
        args: list[any] = None,
    ):
        loop = asyncio.create_task(
            self.run_loop(interval_in_seconds, handles, delay_start, args)
        )
        self.loops.add(loop)
        loop.add_done_callback(self.loops.remove)

    async def run_loop(
        self,
        interval_in_seconds: int,
        handles: Callable | list[Callable],
        delay_start: int = 0,
        args: list[any] = None,
    ):
        if not isinstance(handles, list):
            handles = [handles]

        await asyncio.sleep(delay_start)

        while not self.cancel_request.is_set():
            execution_time = 0
            start_time = time()
            for handle in handles:
                await handle(*args or [])
            execution_time = time() - start_time

            if execution_time < interval_in_seconds:
                await asyncio.sleep(interval_in_seconds - execution_time)
            else:
                await asyncio.sleep(interval_in_seconds)

    async def close(self):
        try:
            await asyncio.wait_for(
                asyncio.gather(
                    *[
                        service.close()
                        for service in [
                            self.influxdb,
                            self.powerflow,
                            self.monitoring,
                            self.weather,
                        ]
                        if service
                    ]
                ),
                timeout=5
            )
        except asyncio.TimeoutError:
            logger.warning(
                "Timeout while closing tasks, proceeding with shutdown.")



from __future__ import annotations

from pydantic import BaseModel, Field, model_validator

from solaredge2mqtt.core.models import EnumModel
from solaredge2mqtt.services.modbus.models.base import ModbusUnitRole


class AdvancedControlsSettings(EnumModel):
    ENABLED = "enabled"
    DISABLED = "disabled"
    DISABLE = "disable"


class ModbusUnitSettings(BaseModel):
    unit: int = Field(1)
    meter: list[bool] = Field(default_factory=list)
    battery: list[bool] = Field(default_factory=list)
    role: ModbusUnitRole = Field(ModbusUnitRole.LEADER, read_only=True)

    @model_validator(mode='before')
    @classmethod
    def fill_defaults(cls, values: dict) -> dict:
        values = cls._fill_defaults_array("meter", values, 3)
        values = cls._fill_defaults_array("battery", values, 2)

        return values

    @staticmethod
    def _fill_defaults_array(
        key: str, values: dict, length: int, default: str = "true"
    ) -> dict:
        if key not in values or not isinstance(values[key], list):
            values[key] = [default] * length
        else:
            for i, value in enumerate(values[key][:length]):
                if isinstance(value, str):
                    values[key][i] = value.lower() == "true"
                elif isinstance(value, bool):
                    values[key][i] = value
                else:
                    values[key][i] = default.lower() == "true"

            if len(values[key]) < length:
                values[key].extend([default] * (length - len(values[key])))

        return values


class ModbusSettings(ModbusUnitSettings):
    host: str = Field(None)
    port: int = Field(1502)

    timeout: int = Field(1)

    check_grid_status: bool = Field(False)
    advanced_power_controls: AdvancedControlsSettings = Field(
        AdvancedControlsSettings.DISABLED)

    follower: list[ModbusUnitSettings] = Field(default_factory=list)

    retain: bool = Field(False)

    @model_validator(mode='before')
    @classmethod
    def fill_defaults(cls, values: dict) -> dict:
        values = super().fill_defaults(values)

        for i, slave_values in enumerate(values.get("follower", [])):
            slave_values["role"] = ModbusUnitRole.FOLLOWER
            slave_values = super()._fill_defaults_array(
                "meter", slave_values, 3, "false"
            )
            slave_values = super()._fill_defaults_array(
                "battery", slave_values, 2, "false"
            )
            values["follower"][i] = slave_values

        return values

    @property
    def advanced_power_controls_enabled(self) -> bool:
        return self.advanced_power_controls == AdvancedControlsSettings.ENABLED

    @property
    def units(self) -> dict[str, ModbusUnitSettings]:
        units = {"leader": self}
        for i, follower in enumerate(self.follower):
            units[f"follower{i}"] = follower

        return units

    @property
    def has_followers(self) -> bool:
        return len(self.follower) > 0



from __future__ import annotations

import base64
import hashlib
import json
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field, computed_field

from solaredge2mqtt.core.models import (
    BaseField,
    BaseInputField,
    BaseInputFieldEnumModel,
    EnumModel,
)

if TYPE_CHECKING:
    from solaredge2mqtt.core.models import BaseInputFieldEnumModel


class HomeAssistantStatus(EnumModel):
    ONLINE = "online"
    OFFLINE = "offline"

    def __init__(self, status: str):
        self._status: str = status

    @property
    def status(self) -> str:
        return self._status


class HomeAssistantStatusInput(BaseInputField):
    status: HomeAssistantStatus

    def __init__(self, status: str):
        super().__init__(status=status)


class HomeAssistantBaseModel(BaseModel):
    client_id: str = Field(..., exclude=True)

    def hash_unique_id(self, ids: list[str | int]) -> str:
        ids.append(self.client_id)
        unique_id = "_".join([str(id) for id in ids])

        hash_obj = hashlib.sha256(unique_id.encode())
        hash_digest = hash_obj.digest()

        base64_encoded = base64.urlsafe_b64encode(hash_digest).decode()

        return base64_encoded[:10]


class HomeAssistantDevice(HomeAssistantBaseModel):
    name: str
    state_topic: str = Field(exclude=True)
    manufacturer: str | None = Field(None)
    model: str | None = Field(None)
    hw_version: str | None = Field(None)
    serial_number: str | None = Field(None)
    sw_version: str | None = Field(None)
    via_device: str | None = Field(None)
    unit_key: str | None = Field(None, exclude=True)

    @computed_field
    @property
    def identifiers(self) -> str:
        identifiers = [self.name, self.manufacturer,
                       self.model, self.serial_number]

        if self.unit_key:
            identifiers.append(self.unit_key)

        return self.hash_unique_id(
            identifiers
        )


class HomeAssistantType(EnumModel):
    BINARY_SENSOR = "binary_sensor", False, []
    NUMBER = "number", True, ["min", "max", "step", "mode"]
    SENSOR = "sensor", False, []

    def __init__(
        self,
        identifier: str,
        command_topic: bool,
        additional_fields: list[str]
    ):
        self._identifier: str = identifier
        self._command_topic: bool = command_topic
        self._additional_fields: list[str] = additional_fields

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def command_topic(self) -> bool:
        return self._command_topic

    @property
    def additional_fields(self) -> list[str]:
        return self._additional_fields


class HomeAssistantEntityBaseType(BaseField):
    def __init__(
        self,
        key: str,
        typed: HomeAssistantType,
        device_class: str | None = None,
        state_class: str | None = None,
        unit_of_measurement: str | None = None,
    ):
        super().__init__(key)

        self._typed: HomeAssistantType = typed
        self._device_class: str | None = device_class
        self._state_class: str | None = state_class
        self._unit_of_measurement: str | None = unit_of_measurement

    @property
    def typed(self) -> HomeAssistantType:
        return self._typed

    @property
    def device_class(self) -> str | None:
        return self._device_class

    @property
    def state_class(self) -> str | None:
        return self._state_class

    @property
    def unit_of_measurement(self) -> str | None:
        return self._unit_of_measurement

    def field(
        self,
        title: str | None = None,
        icon: str | None = None,
        json_schema_extra: dict[str, any] | None = None,
        input_field: BaseInputFieldEnumModel | None = None
    ) -> dict[str, any]:
        if json_schema_extra is None:
            json_schema_extra = {}

        json_schema_extra = {
            "ha_type": self,
            "ha_typed": self.typed.identifier,
            "icon": icon,
            **json_schema_extra
        }

        return super().field(title, json_schema_extra, input_field)


class HomeAssistantBinarySensorType(HomeAssistantEntityBaseType):
    ENABLED = "enabled", None, None
    GRID_STATUS = "grid_status", "power", None
    PLUG = "plug", "plug", None

    def __init__(
        self,
        key: str,
        device_class: str | None = None,
        state_class: str | None = None,
    ):
        super().__init__(
            key,
            HomeAssistantType.BINARY_SENSOR,
            device_class,
            state_class,
            None,
        )


class HomeAssistantNumberType(HomeAssistantEntityBaseType):
    ACTIVE_POWER_LIMIT = "active_power_limit", None, '%', 0, 100, 1, "slider"

    def __init__(
        self,
        key: str,
        device_class: str | None = None,
        unit_of_measurement: str | None = None,
        min: int | float | None = None,
        max: int | float | None = None,
        step: int | float | None = None,
        mode: str | None = None,
    ):
        super().__init__(key, HomeAssistantType.NUMBER,
                         device_class, None, unit_of_measurement)

        self._min: int | float | None = min
        self._max: int | float | None = max
        self._step: int | float | None = step
        self._mode: str | None = mode

    def field(
        self,
        input_field: BaseInputFieldEnumModel,
        title: str | None = None,
        icon: str | None = None,
        min: float | None = None,
        max: float | None = None,
        step: float | None = None,
        mode: str | None = None
    ) -> dict[str, any]:
        json_schema_extra = {
            "min": min or self._min,
            "max": max or self._max,
            "step": step or self._step,
            "mode": mode or self._mode,
        }

        return super().field(title, icon, json_schema_extra, input_field)


class HomeAssistantSensorType(HomeAssistantEntityBaseType):
    APPARENT_POWER = "apparent_power", "apparent_power", "measurement", "VA"
    BATTERY = "battery", "battery", "measurement", "%"
    CURRENT_A = "current_a", "current", "measurement", "A"
    ENERGY_KWH = "energy_kwh", "energy", "total_increasing", "kWh"
    ENERGY_WH = "energy_wh", "energy", "total_increasing", "Wh"
    FREQUENCY_HZ = "frequency_hz", "frequency", "measurement", "Hz"
    MONETARY = "monetary", "monetary", "total", None
    MONETARY_BALANCE = "monetary_balance", "monetary", None, None
    PERCENTAGE = "percentage", None, "measurement", "%"
    POWER_FACTOR = "power_factor", "power_factor", "measurement", "%"
    POWER_KW = "power_kw", "power", "measurement", "kW"
    POWER_W = "power_w", "power", "measurement", "W"
    REACTIVE_POWER = "reactive_power", "reactive_power", "measurement", "var"
    STATUS = "status", None, None, None
    TEMP_C = "temp_c", "temperature", "measurement", "°C"
    VOLTAGE_V = "voltage_v", "voltage", "measurement", "V"

    def __init__(
        self,
        key: str,
        device_class: str | None = None,
        state_class: str | None = None,
        unit_of_measurement: str | None = None,
    ):
        super().__init__(key, HomeAssistantType.SENSOR,
                         device_class, state_class, unit_of_measurement)


class HomeAssistantEntity(HomeAssistantBaseModel):
    name: str
    device: HomeAssistantDevice
    _icon: str | None = None
    _additional_fields: dict[str, any] = {}
    path: list[str] | None = Field(None, exclude=True)
    ha_type: HomeAssistantEntityBaseType = Field(
        exclude=True)
    unit: str | None = Field(None, exclude=True)

    def __init__(self, device: HomeAssistantDevice, icon: str | None = None, **data):
        super().__init__(client_id=device.client_id, device=device, **data)

        for field in self.ha_type.typed.additional_fields:
            if field in data and data[field] is not None:
                self._additional_fields[field] = data[field]

        self._icon = icon

    @computed_field
    @property
    def unique_id(self) -> str:
        return self.hash_unique_id(
            [self.device.identifiers, self.name,
                self.state_topic, self.value_template]
        )

    @computed_field
    @property
    def command_topic(self) -> str | None:
        return (
            f"{self.device.state_topic}/{'/'.join(self.path)}"
            if self.ha_type.typed.command_topic
            else None
        )

    @computed_field
    @property
    def state_topic(self) -> str:
        return self.device.state_topic

    @computed_field
    @property
    def value_template(self) -> str | None:
        return f"{{{{ value_json.{'.'.join(self.path)} }}}}" if self.path else None

    @computed_field
    @property
    def state_class(self) -> str | None:
        return self.ha_type.state_class

    @computed_field
    @property
    def device_class(self) -> str | None:
        return self.ha_type.device_class

    @computed_field
    @property
    def unit_of_measurement(self) -> str | None:
        return self.ha_type.unit_of_measurement if self.unit is None else self.unit

    @computed_field
    @property
    def payload_on(self) -> bool | None:
        return True if self.ha_type.typed == HomeAssistantType.BINARY_SENSOR else None

    @computed_field
    @property
    def payload_off(self) -> bool | None:
        return False if self.ha_type.typed == HomeAssistantType.BINARY_SENSOR else None

    @computed_field
    @property
    def icon(self) -> str | None:
        return f"mdi:{self._icon}" if self._icon else None

    def model_dump_json(self, **kwargs) -> str:
        dumped_model = super().model_dump(**kwargs)
        dumped_model = {**dumped_model, **self._additional_fields}
        return json.dumps(dumped_model)



from __future__ import annotations

from pymodbus.client import ModbusTcpClient

from solaredge2mqtt.core.models import EnumModel
from solaredge2mqtt.services.modbus.sunspec.values import (
    SunSpecInputData,
    SunSpecPayload,
    SunSpecValueType,
)


class SunSpecRequestRegisterBundle:
    @classmethod
    def from_registers(
        cls, registers: list[SunSpecRegister], required_only: bool = True
    ) -> list[SunSpecRequestRegisterBundle]:

        sorted_registers = sorted(registers, key=lambda reg: reg.address)

        bundles = cls._bundle_registers(required_only, sorted_registers)

        if required_only:
            bundles = cls._add_not_required(sorted_registers, bundles)

        return bundles

    @classmethod
    def _bundle_registers(
        cls, required_only, sorted_registers
    ) -> list[SunSpecRequestRegisterBundle]:
        bundles = []
        current_bundle = cls()

        for register in [
            reg for reg in sorted_registers if not required_only or reg.required
        ]:
            if (
                current_bundle.length > 0
                and register.end_address - current_bundle.address + 1 > 120
            ):
                bundles.append(current_bundle)
                current_bundle = cls()

            current_bundle.add_register(register)

        if current_bundle.length > 0:
            bundles.append(current_bundle)

        return bundles

    @staticmethod
    def _add_not_required(
        sorted_registers, bundles
    ) -> list[SunSpecRequestRegisterBundle]:
        for bundle in bundles:
            not_required_registers = [
                reg
                for reg in sorted_registers
                if not reg.required
                and reg.address >= bundle.address
                and reg.end_address <= bundle.end_address
            ]

            for reg in not_required_registers:
                bundle.add_register(reg)

        return bundles

    def __init__(self):
        self._registers: list[SunSpecRegister] = []

    def add_register(self, register: SunSpecRegister) -> None:
        self._registers.append(register)

    @property
    def registers(self) -> set[SunSpecRegister]:
        return self._registers

    @property
    def address(self) -> int:
        return min([register.address for register in self._registers])

    @property
    def end_address(self) -> int:
        return max([register.end_address for register in self._registers])

    @property
    def length(self) -> int:
        length = self.end_address - self.address if self._registers else 0

        return length

    def decode_response(
        self, registers: list[int], data: dict[str, SunSpecPayload]
    ) -> dict[str, SunSpecPayload]:
        for register in self._registers:
            offset = register.address - self.address
            response_slice = registers[offset: offset + register.length]
            data = register.decode_response(response_slice, data)

        return data


class SunSpecRegister(EnumModel):
    def __init__(
        self,
        identifier: str,
        address: int,
        value_type: SunSpecValueType,
        required: bool = False,
        length: int = 0,
    ) -> None:
        # pylint: disable=super-init-not-called
        self._identifier = identifier
        self._address = address
        self._value_type = value_type

        if length != 0:
            self._length = length
        elif value_type.length != 0:
            self._length = value_type.length

        self._required = required

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def address(self) -> int:
        return self._address

    @property
    def end_address(self) -> int:
        return self.address + self.length

    @property
    def value_type(self) -> SunSpecValueType:
        return self._value_type

    @property
    def length(self) -> int:
        return self._length

    @property
    def required(self) -> bool:
        return self._required

    @staticmethod
    def wordorder() -> str:
        return "big"

    @classmethod
    def request_bundles(
        cls, required_only: bool = True
    ) -> list[SunSpecRequestRegisterBundle]:
        if not hasattr(cls, "_cached_bundles"):
            cls._cached_bundles = SunSpecRequestRegisterBundle.from_registers(
                cls, required_only
            )
        return cls._cached_bundles

    def decode_response(
        self, registers: list[int], data: dict[str, SunSpecPayload]
    ) -> dict[str, SunSpecPayload]:
        value = ModbusTcpClient.convert_from_registers(
            registers, self.value_type.data_type, word_order=self.wordorder()
        )

        if self.value_type == SunSpecValueType.STRING:
            value = value.strip("\x00").rstrip()
        elif value == self.value_type.not_implemented_value:
            value = False

        data[self.identifier] = value

        return data

    def encode_request(self, value: SunSpecInputData) -> list[int]:
        if isinstance(value, bool):
            value = 1 if value else 0

        value = ModbusTcpClient.convert_to_registers(
            value, self.value_type.data_type, word_order=self.wordorder()
        )

        return value


class SunSpecOffset(EnumModel):
    def __init__(self, identifier: str, offset: int):
        # pylint: disable=super-init-not-called
        self._identifier = identifier
        self._offset = offset

    @property
    def idx(self) -> int:
        return int(self._identifier[-1])

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def offset(self) -> int:
        return self._offset



from pydantic import BaseModel, Field, SecretStr

SECONDS_PER_DAY = 86400
SECONDS_PER_HOUR = 3600
SECONDS_PER_YEAR = SECONDS_PER_DAY * 365
SECONDS_PER_2_YEARS = SECONDS_PER_YEAR * 2


class InfluxDBSettings(BaseModel):
    host: str = Field(None)
    port: int = Field(8086)
    token: SecretStr = Field(None)
    org: str = Field(None)
    bucket: str = Field("solaredge")
    retention: int = Field(SECONDS_PER_2_YEARS)
    retention_raw: int = Field(25)

    @property
    def url(self) -> str:
        url = f"{self.host}:{self.port}"
        if not str(self.host).startswith(("http://", "https://")):
            url = f"http://{url}"

        return url

    @property
    def client_params(self) -> dict[str, any]:
        return {
            "url": self.url,
            "token": self.token.get_secret_value(),
            "org": self.org,
        }

    @property
    def is_configured(self) -> bool:
        return all(
            [
                self.host is not None,
                self.port is not None,
                self.token is not None,
                self.org is not None,
            ]
        )



from solaredge2mqtt.core.events.events import BaseEvent
from solaredge2mqtt.services.weather.models import OpenWeatherMapOneCall


class WeatherUpdateEvent(BaseEvent):
    def __init__(self, weather: OpenWeatherMapOneCall) -> None:
        self._weather = weather

    @property
    def weather(self) -> OpenWeatherMapOneCall:
        return self._weather



from solaredge2mqtt.core.events.events import BaseEvent
from solaredge2mqtt.services.models import Component


class ComponentEvent(BaseEvent):
    def __init__(self, component: Component):
        self._component = component

    @property
    def component(self) -> Component:
        return self._component

    def __str__(self) -> str:
        return str(self._component)


class ComponentsEvent(BaseEvent):
    def __init__(self, components: dict[str, Component]):
        self._components = components

    @property
    def components(self) -> dict[str, Component]:
        return self._components

    def __str__(self) -> str:
        return str(self._components)



from __future__ import annotations

from solaredge2mqtt.core.models import EnumModel, Solaredge2MQTTBaseModel
from solaredge2mqtt.services.modbus.sunspec.values import C_SUNSPEC_DID_MAP
from solaredge2mqtt.services.models import Component


class ModbusUnitRole(EnumModel):
    LEADER = "leader"
    FOLLOWER = "follower"
    CUMULATED = "cumulated"

    def __init__(self, role: str):
        self._role: str = role

    @property
    def role(self) -> str:
        return self._role


class ModbusUnitInfo(Solaredge2MQTTBaseModel):
    unit: int
    key: str
    role: ModbusUnitRole


class ModbusDeviceInfo(Solaredge2MQTTBaseModel):
    manufacturer: str
    model: str
    option: str | None = None
    sunspec_type: str
    version: str
    serialnumber: str
    unit: ModbusUnitInfo | None = None

    def __init__(self, data: dict[str, str | int]):
        values = {
            "manufacturer": data["c_manufacturer"],
            "model": data["c_model"],
            "version": data["c_version"],
            "serialnumber": data["c_serialnumber"]
        }

        if "unit" in data:
            values["unit"] = data["unit"]

        if "c_sunspec_did" in data and data["c_sunspec_did"] in C_SUNSPEC_DID_MAP:
            values["sunspec_type"] = C_SUNSPEC_DID_MAP[data["c_sunspec_did"]]
        else:
            values["sunspec_type"] = "Unknown"

        if "c_option" in data:
            values["option"] = data["c_option"]

        super().__init__(**values)

    def homeassistant_device_info(self, name: str) -> dict[str, any]:
        info = {
            "name": f"SolarEdge {name}",
            "manufacturer": self.manufacturer,
            "model": self.model,
            "hw_version": self.version,
            "serial_number": self.serialnumber
        }

        if self.unit:
            info["unit_key"] = self.unit.key

        return info

    @property
    def has_unit(self) -> bool:
        return self.unit is not None

    def unit_key(self, suffix: str = "") -> str:
        return f"{self.unit.key}{suffix}" if self.has_unit else ""


class ModbusComponent(Component):
    SOURCE = "modbus"

    info: ModbusDeviceInfo

    def model_dump_influxdb(self, exclude: list[str] | None = None) -> dict[str, any]:
        return super().model_dump_influxdb(["info", *exclude] if exclude else ["info"])

    @property
    def influxdb_tags(self) -> dict[str, str]:
        return {
            **super().influxdb_tags,
            "manufacturer": self.info.manufacturer,
            "model": self.info.model,
            "option": self.info.option,
            "sunspec_type": self.info.sunspec_type,
            "serialnumber": self.info.serialnumber,
        }

    @classmethod
    # pylint: disable=arguments-differ
    def model_json_schema(cls, mode: str = "serialization") -> dict[str, any]:
        schema = super().model_json_schema(mode=mode)
        schema["properties"].pop("info", None)
        return schema

    def homeassistant_device_info(self) -> dict[str, any]:
        return self.info.homeassistant_device_info()

    def mqtt_topic(self, has_followers: bool = False) -> str:
        return self.generate_topic_prefix(self.info.unit.key if has_followers else None)

    @classmethod
    def generate_topic_prefix(cls, unit_key: str | None = None) -> str:
        topic_parts = [cls.SOURCE]

        if unit_key:
            topic_parts.append(unit_key)

        topic_parts.append(cls.COMPONENT)

        return "/".join(topic_parts)

    @property
    def has_unit(self) -> bool:
        return self.info.has_unit



from __future__ import annotations

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import InvalidDataException
from solaredge2mqtt.core.influxdb import InfluxDBAsync
from solaredge2mqtt.core.influxdb.events import InfluxDBAggregatedEvent
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTPublishEvent
from solaredge2mqtt.services.energy.events import EnergyReadEvent
from solaredge2mqtt.services.energy.models import (
    HistoricEnergy,
    HistoricPeriod,
    HistoricQuery,
)
from solaredge2mqtt.services.energy.settings import EnergySettings


class EnergyService:
    def __init__(
        self,
        settings: EnergySettings,
        event_bus: EventBus,
        influxdb: InfluxDBAsync,
    ):
        self.influxdb = influxdb
        self.settings = settings

        self.event_bus = event_bus
        self._subscribe_events()

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(
            InfluxDBAggregatedEvent, self.read_historic_energy)

    async def read_historic_energy(self, _) -> None:
        for period in HistoricPeriod:
            records = await self.influxdb.query_timeunit(period, "energy")
            if records is None:
                if period.query == HistoricQuery.LAST:
                    logger.info(
                        "No data found for {period}, skipping this loop", period=period
                    )
                else:
                    raise InvalidDataException(f"No energy data for {period}")

                continue

            for record in records:
                energy = HistoricEnergy(record, period)

                logger.info(
                    "Read from influxdb {period} energy: {energy.pv_production} kWh",
                    period=period,
                    energy=energy,
                )

                await self.event_bus.emit(EnergyReadEvent(energy))
                await self.event_bus.emit(
                    MQTTPublishEvent(
                        energy.mqtt_topic(),
                        energy,
                        self.settings.retain,
                    )
                )



from datetime import datetime

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.timer.events import (
    Interval1MinTriggerEvent,
    Interval5MinTriggerEvent,
    Interval10MinTriggerEvent,
    Interval15MinTriggerEvent,
    IntervalBaseTriggerEvent,
)


class Timer:
    def __init__(self, event_bus: EventBus, base_interval: int) -> None:
        self.event_bus = event_bus
        self.base_interval = base_interval

    async def loop(self) -> None:
        timestamp = int(datetime.now().timestamp())

        if timestamp % self.base_interval == 0:
            await self.event_bus.emit(IntervalBaseTriggerEvent())

        timestamp -= self.base_interval - 1

        if timestamp % 60 == 0:
            await self.event_bus.emit(Interval1MinTriggerEvent())

        timestamp -= self.base_interval

        if timestamp % 300 == 0:
            await self.event_bus.emit(Interval5MinTriggerEvent())

        timestamp -= self.base_interval

        if timestamp % 600 == 0:
            await self.event_bus.emit(Interval10MinTriggerEvent())
        timestamp -= self.base_interval

        if timestamp % 900 == 0:
            await self.event_bus.emit(Interval15MinTriggerEvent)



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from importlib import resources
from typing import TYPE_CHECKING

from influxdb_client import BucketRetentionRules, InfluxDBClient, Point
from influxdb_client.client.bucket_api import BucketsApi
from influxdb_client.client.delete_api import DeleteApi
from influxdb_client.client.influxdb_client_async import InfluxDBClientAsync
from influxdb_client.client.query_api_async import QueryApiAsync
from tzlocal import get_localzone_name

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.influxdb.events import InfluxDBAggregatedEvent
from solaredge2mqtt.core.influxdb.settings import InfluxDBSettings
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.timer.events import Interval10MinTriggerEvent

if TYPE_CHECKING:
    from pandas import DataFrame

    from solaredge2mqtt.services.energy.models import HistoricPeriod
    from solaredge2mqtt.services.energy.settings import PriceSettings

LOCAL_TZ = get_localzone_name()


class InfluxDBAsync:
    def __init__(
        self,
        settings: InfluxDBSettings,
        prices: PriceSettings,
        event_bus: EventBus | None = None,
    ) -> None:
        self.settings: InfluxDBSettings = settings
        self.prices: PriceSettings = prices

        self.event_bus = event_bus
        if self.event_bus:
            self._subscribe_events()

        self.client_async: InfluxDBClientAsync | None = None
        self.client_sync: InfluxDBClient = InfluxDBClient(
            **self.settings.client_params)

        self.flux_cache: dict[str, str] = {}

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(Interval10MinTriggerEvent, self.loop)

    def init(self) -> None:
        self.client_async = InfluxDBClientAsync(**self.settings.client_params)

        self.initialize_buckets()

    def initialize_buckets(self) -> None:
        bucket = self.buckets_api.find_bucket_by_name(self.bucket_name)
        retention_rules = BucketRetentionRules(
            type="expire", every_seconds=self.settings.retention
        )

        if bucket is None:
            logger.info(f"Creating bucket '{self.bucket_name}'")
            self.buckets_api.create_bucket(
                bucket_name=self.bucket_name, retention_rules=retention_rules
            )
        else:
            logger.info(f"Bucket '{self.bucket_name}' already exists.")
            if bucket.retention_rules[0].every_seconds != self.settings.retention:
                logger.info(
                    f"Updating retention rules for bucket '{self.bucket_name}' "
                    + f"to {self.settings.retention} seconds."
                )
                bucket.retention_rules[0] = retention_rules
                self.buckets_api.update_bucket(bucket=bucket)

    @property
    def buckets_api(self) -> BucketsApi:
        return self.client_sync.buckets_api()

    async def loop(self, _) -> None:
        now = datetime.now(tz=timezone.utc).replace(
            minute=0, second=0, microsecond=0)

        logger.info("Aggregate powerflow and energy raw data")
        aggregate_query = self._get_flux_query(
            "aggregate",
            {"PRICE_IN": self.prices.price_in, "PRICE_OUT": self.prices.price_out},
        )
        await self.query_api.query(aggregate_query)

        logger.info("Apply retention on raw data")
        retention_time = now - timedelta(hours=self.settings.retention_raw)
        self.delete_from_measurements(
            datetime(1970, 1, 1, tzinfo=timezone.utc),
            retention_time,
            ["powerflow_raw", "battery_raw"],
        )

        if self.event_bus:
            await self.event_bus.emit(InfluxDBAggregatedEvent())

    @property
    def query_api(self) -> QueryApiAsync:
        return self.client_async.query_api()

    def delete_from_measurements(
        self, start: datetime, stop: datetime, measurements: list[str]
    ) -> None:
        for measurement in measurements:
            self.delete_from_measurement(start, stop, measurement)

    def delete_from_measurement(
        self, start: datetime, stop: datetime, measurement: str
    ) -> None:
        self.delete_api.delete(
            start, stop, f'_measurement="{measurement}"', self.bucket_name
        )

    @property
    def delete_api(self) -> DeleteApi:
        return self.client_sync.delete_api()

    @property
    def bucket_name(self) -> str:
        return f"{self.settings.bucket}"

    async def write_point(self, point: Point) -> None:
        await self.write_points([point])

    async def write_points(self, points: list[Point]) -> None:
        await self.client_async.write_api().write(
            bucket=self.bucket_name, record=points
        )

    async def query_timeunit(
        self, period: HistoricPeriod, measurement: str
    ) -> list[dict[str, any]] | None:
        results = await self.query(
            period.query.query, {"UNIT": period.unit,
                                 "MEASUREMENT": measurement}
        )

        return results if len(results) > 0 else None

    async def query_first(
        self, query_name: str, additional_replacements: dict[str, any] | None = None
    ) -> dict[str, any] | None:
        result = await self.query(query_name, additional_replacements)
        return result[0] if len(result) > 0 else None

    async def query(
        self, query_name: str, additional_replacements: dict[str, any] | None = None
    ) -> list[dict[str, any]]:

        tables = await self.query_api.query(
            self._get_flux_query(query_name, additional_replacements)
        )
        return [record.values for table in tables for record in table.records]

    async def query_dataframe(
        self, query_name: str, additional_replacements: dict[str, any] | None = None
    ) -> DataFrame:
        return await self.query_api.query_data_frame(
            self._get_flux_query(query_name, additional_replacements)
        )

    def _get_flux_query(
        self, query_name: str, additional_replacements: dict[str, any] | None = None
    ) -> str:
        if query_name not in self.flux_cache:
            with resources.files(__package__).joinpath(
                f"flux/{query_name}.flux"
            ).open("r", encoding="utf-8") as f:
                flux = f.read()

            flux = (
                flux.replace("{{BUCKET_AGGREGATED}}", self.bucket_name)
                .replace("{{BUCKET_NAME}}", self.bucket_name)
                .replace("{{TIMEZONE}}", LOCAL_TZ)
            )
            self.flux_cache[query_name] = flux

        query = self.flux_cache[query_name]

        if additional_replacements is not None:
            for key, value in additional_replacements.items():
                query = query.replace("{{" + key + "}}", str(value))

        logger.trace(query)

        return query

    async def close(self) -> None:
        if self.client_async:
            await self.client_async.close()
            self.client_async = None



from pydantic import Field

from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantSensorType as HASensor,
)
from solaredge2mqtt.services.models import ComponentValueGroup


class ModbusACCurrent(ComponentValueGroup):
    actual: float = Field(**HASensor.CURRENT_A.field("actual"))
    l1: float | None = Field(None, **HASensor.CURRENT_A.field("L1"))
    l2: float | None = Field(None, **HASensor.CURRENT_A.field("L2"))
    l3: float | None = Field(None, **HASensor.CURRENT_A.field("L3"))

    def __init__(self, data: dict[str, str | int]):
        values = {"actual": self.scale_value(data, "current")}

        for phase in ["l1", "l2", "l3"]:
            if f"{phase}_current" in data:
                values[phase] = self.scale_value(
                    data, f"{phase}_current", "current_scale"
                )

        super().__init__(**values)


class ModbusACVoltage(ComponentValueGroup):
    l1: float | None = Field(None, **HASensor.VOLTAGE_V.field("L1"))
    l2: float | None = Field(None, **HASensor.VOLTAGE_V.field("L2"))
    l3: float | None = Field(None, **HASensor.VOLTAGE_V.field("L3"))
    l1n: float | None = Field(None, **HASensor.VOLTAGE_V.field("L1N"))
    l2n: float | None = Field(None, **HASensor.VOLTAGE_V.field("L2N"))
    l3n: float | None = Field(None, **HASensor.VOLTAGE_V.field("L3N"))

    def __init__(self, data: dict[str, str | int]):
        values = {}
        for phase in ["l1", "l2", "l3"]:
            if f"{phase}_voltage" in data:
                values[phase] = self.scale_value(
                    data, f"{phase}_voltage", "voltage_scale"
                )

            if f"{phase}n_voltage" in data:
                values[f"{phase}n"] = self.scale_value(
                    data, f"{phase}n_voltage", "voltage_scale"
                )

        super().__init__(**values)


class ModbusACPower(ComponentValueGroup):
    actual: float = Field(**HASensor.POWER_W.field("actual"))
    reactive: float = Field(**HASensor.REACTIVE_POWER.field("reactive"))
    apparent: float = Field(**HASensor.APPARENT_POWER.field("apparent"))
    factor: float = Field(**HASensor.POWER_FACTOR.field("power factor"))

    def __init__(self, data: dict[str, str | int], power_key: str):
        actual = self.scale_value(data, power_key)
        reactive = self.scale_value(data, "power_reactive")
        apparent = self.scale_value(data, "power_apparent")
        factor = self.scale_value(data, "power_factor")
        super().__init__(
            actual=actual, reactive=reactive, apparent=apparent, factor=factor
        )


class ModbusAC(ComponentValueGroup):
    current: ModbusACCurrent = Field(title="Current")
    voltage: ModbusACVoltage = Field(title="Voltage")
    power: ModbusACPower = Field(title="Power")
    frequency: float = Field(**HASensor.FREQUENCY_HZ.field("Grid frequency"))

    def __init__(self, data: dict[str, str | int]):
        current = ModbusACCurrent(data)
        voltage = ModbusACVoltage(data)
        power = ModbusACPower(data, "power_ac")
        frequency = self.scale_value(data, "frequency")

        super().__init__(
            current=current, voltage=voltage, power=power, frequency=frequency
        )


class ModbusEnergy(ComponentValueGroup):
    totalexport: float = Field(**HASensor.ENERGY_WH.field("Export"))
    totalimport: float = Field(**HASensor.ENERGY_WH.field("Import"))

    def __init__(self, data: dict[str, str | int]):
        super().__init__(
            totalexport=self.scale_value(
                data, "export_energy_active", "energy_active_scale"
            ),
            totalimport=self.scale_value(
                data, "import_energy_active", "energy_active_scale"
            ),
        )


class ModbusDC(ComponentValueGroup):
    current: float = Field(**HASensor.CURRENT_A.field("current"))
    voltage: float = Field(**HASensor.VOLTAGE_V.field("voltage"))
    power: float = Field(**HASensor.POWER_W.field("power"))

    def __init__(self, data: dict[str, str | int]):
        super().__init__(
            current=self.scale_value(data, "current_dc"),
            voltage=self.scale_value(data, "voltage_dc"),
            power=self.scale_value(data, "power_dc"),
        )



from __future__ import annotations

import asyncio
import json
from typing import TYPE_CHECKING

from pymodbus.client import AsyncModbusTcpClient
from pymodbus.exceptions import ModbusException

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import InvalidDataException
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.services.modbus.control import ModbusAdvancedControl
from solaredge2mqtt.services.modbus.events import ModbusUnitsReadEvent, ModbusWriteEvent
from solaredge2mqtt.services.modbus.models.base import ModbusDeviceInfo
from solaredge2mqtt.services.modbus.models.battery import ModbusBattery
from solaredge2mqtt.services.modbus.models.inverter import ModbusInverter
from solaredge2mqtt.services.modbus.models.meter import ModbusMeter
from solaredge2mqtt.services.modbus.models.unit import ModbusUnit
from solaredge2mqtt.services.modbus.settings import ModbusUnitSettings
from solaredge2mqtt.services.modbus.sunspec.base import (
    SunSpecRegister,
    SunSpecRequestRegisterBundle,
)
from solaredge2mqtt.services.modbus.sunspec.battery import (
    SunSpecBatteryInfoRegister,
    SunSpecBatteryOffset,
    SunSpecBatteryRegister,
)
from solaredge2mqtt.services.modbus.sunspec.inverter import (
    SunSpecGridStatusRegister,
    SunSpecInverterInfoRegister,
    SunSpecInverterRegister,
    SunSpecPowerControlRegister,
)
from solaredge2mqtt.services.modbus.sunspec.meter import (
    SunSpecMeterInfoRegister,
    SunSpecMeterOffset,
    SunSpecMeterRegister,
)
from solaredge2mqtt.services.modbus.sunspec.values import (
    SunSpecInputData,
    SunSpecPayload,
)

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings.models import ServiceSettings

LOGGING_DEVICE_INFO = (
    "{unit_key}{device} ({info.manufacturer} {info.model} {info.serialnumber})"
)


class Modbus:
    def __init__(self, settings: ServiceSettings, event_bus: EventBus):
        self.settings = settings.modbus

        logger.info(
            "Using SolarEdge inverter via modbus: {host}:{port}",
            host=self.settings.host,
            port=self.settings.port,
        )

        logger.debug(f"Modbus settings: {self.settings}")

        self.event_bus = event_bus

        self._block_unreadable: set[int] = set()

        self._initialized = False
        self._device_info: dict[str, dict[str, ModbusDeviceInfo]] = {}

        self.client: AsyncModbusTcpClient | None = None

        self._control: ModbusAdvancedControl = ModbusAdvancedControl(
            settings, event_bus)

        self._subscribe_events()

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(ModbusWriteEvent, self._handle_write_event)

    async def async_init(self) -> None:
        logger.info("Initializing modbus")

        self.client = AsyncModbusTcpClient(
            host=self.settings.host,
            port=self.settings.port,
            timeout=self.settings.timeout,
            retries=0,
        )

        await self.detect_devices()

        await asyncio.sleep(self.settings.timeout + 5)

        await self.check_readable_registers()

        self._initialized = True

        await asyncio.sleep(self.settings.timeout + 5)
        if self._block_unreadable:
            logger.warning(
                "Not readable registers: {registers}", registers=self._block_unreadable
            )

        # await self._control.async_init()

    async def detect_devices(self):
        async with self.client:
            for unit_key, unit_settings in self.settings.units.items():
                logger.info(f"Detecting devices for unit: {unit_key}")

                self._device_info[unit_key] = {}

                inverter_raw = await self.read_device_info(
                    SunSpecInverterInfoRegister, unit_key, "inverter", unit_settings
                )

                logger.debug(f"Detected inverter: {inverter_raw}")

                for meter in SunSpecMeterOffset:
                    if (
                        unit_settings.meter[meter.idx]
                        and meter.identifier in inverter_raw
                        and inverter_raw[meter.identifier] > 0
                    ):
                        await self.read_device_info(
                            SunSpecMeterInfoRegister,
                            unit_key,
                            meter.identifier,
                            unit_settings,
                            meter.offset,
                        )

                for battery in SunSpecBatteryOffset:
                    if (
                        unit_settings.battery[battery.idx]
                        and battery.identifier in inverter_raw
                        and inverter_raw[battery.identifier] != 255
                    ):
                        await self.read_device_info(
                            SunSpecBatteryInfoRegister,
                            unit_key,
                            battery.identifier,
                            unit_settings,
                            battery.offset,
                        )

    async def read_device_info(
        self,
        registers: SunSpecRegister,
        unit_key: str,
        key: str,
        unit_settings: ModbusUnitSettings,
        offset: int = 0
    ) -> SunSpecPayload:
        raw_data = await self._read_from_modbus(registers,
                                                unit_settings.unit,
                                                offset)

        if self.settings.has_followers:
            raw_data["unit"] = {
                "unit": unit_settings.unit,
                "key": unit_key,
                "role": unit_settings.role
            }

        info = ModbusDeviceInfo(raw_data)
        logger.info(
            f"Found {key} {info.manufacturer} {info.model} {info.serialnumber}")
        self._device_info[unit_key][key] = info
        return raw_data

    async def check_readable_registers(self):
        self.client = AsyncModbusTcpClient(
            host=self.settings.host,
            port=self.settings.port,
            timeout=self.settings.timeout,
            retries=1,
        )

        async with self.client:
            for unit_key, unit_settings in self.settings.units.items():
                logger.info(
                    f"Checking modbus registers from unit: {unit_key}",)
                await self._get_raw_data(unit_key, unit_settings.unit)

    async def get_data(
        self,
    ) -> ModbusUnit:
        units: dict[str, ModbusUnit] = {}

        try:
            async with self.client:
                for unit_key, unit_settings in self.settings.units.items():
                    inverter_raw, meters_raw, batteries_raw = await self._get_raw_data(
                        unit_key, unit_settings.unit
                    )

                    inverter_data = self._map_inverter(unit_key, inverter_raw)
                    meters_data = self._map_meters(unit_key, meters_raw)
                    batteries_data = self._map_batteries(unit_key, batteries_raw)
    
                    units[unit_key] = ModbusUnit(
                        info=inverter_data.info.unit,
                        inverter=inverter_data,
                        meters=meters_data,
                        batteries=batteries_data
                    )

        except KeyError as error:
            raise InvalidDataException("Invalid modbus data") from error

        await self.event_bus.emit(ModbusUnitsReadEvent(units))

        return units

    async def _get_raw_data(
        self,
        unit_key: str,
        unit: int
    ) -> tuple[SunSpecPayload, dict[str, SunSpecPayload], dict[str, SunSpecPayload]]:

        inverter_raw = await self._read_from_modbus(
            SunSpecInverterRegister.request_bundles(),
            unit,
        )

        if self.settings.check_grid_status:
            grid_status_raw = await self._read_from_modbus(
                SunSpecGridStatusRegister.request_bundles(),
                unit,
            )
            inverter_raw = {**inverter_raw, **grid_status_raw}

        if self.settings.advanced_power_controls_enabled:
            advanced_power_control_raw = await self._read_from_modbus(
                SunSpecPowerControlRegister.request_bundles(),
                unit,
            )
            inverter_raw = {
                **inverter_raw,
                **advanced_power_control_raw,
            }

            logger.debug(advanced_power_control_raw)

        meters_raw = {}
        batteries_raw = {}

        for meter in SunSpecMeterOffset:
            if meter.identifier in self._device_info[unit_key]:
                meters_raw[meter.identifier] = await self._read_from_modbus(
                    SunSpecMeterRegister.request_bundles(),
                    unit,
                    offset=meter.offset
                )

        for battery in SunSpecBatteryOffset:
            if battery.identifier in self._device_info[unit_key]:
                batteries_raw[battery.identifier] = await self._read_from_modbus(
                    SunSpecBatteryRegister.request_bundles(),
                    unit,
                    offset=battery.offset
                )

        return inverter_raw, meters_raw, batteries_raw

    async def _read_from_modbus(
        self,
        registers_or_bundles: SunSpecRegister | list[SunSpecRequestRegisterBundle],
        unit: int,
        offset: int = 0,
    ) -> SunSpecPayload:
        data = {}

        for register_or_bundle in registers_or_bundles:
            address_start = register_or_bundle.address + offset

            if address_start in self._block_unreadable:
                logger.trace(
                    f"Skip unreadable registers beginning at {address_start}")
                continue

            logger.trace(
                (
                    f"Read {register_or_bundle.length} register "
                    f"beginning at {address_start}"
                )
            )

            try:
                result = await self.client.read_holding_registers(
                    device_id=unit,
                    address=address_start,
                    count=register_or_bundle.length,
                )

                if result.isError():
                    logger.error(f"Unreadable register {address_start}")
                    logger.debug(f"Modbus read error: {result}")
                    self._block_register(address_start)
                else:
                    data = register_or_bundle.decode_response(
                        result.registers, data)

                if not self._initialized:
                    logger.trace(
                        (
                            f"Checked {register_or_bundle.length} registers "
                            f"from {address_start} successfully"
                        )
                    )

            except ModbusException as error:
                logger.debug(f"Modbus read exception: {error}")
                logger.error(f"Unreadable register {address_start}")
                self._block_register(address_start)

        return data

    def _block_register(self, register: int) -> None:
        if not self._initialized:
            logger.info(f"Block unreadable register beginning at {register}")
            self._block_unreadable.add(register)

    def _map_inverter(
        self, unit_key: str, inverter_raw: SunSpecPayload
    ) -> ModbusInverter:
        logger.debug(
            "Inverter raw:\n{raw}",
            raw=json.dumps(inverter_raw, indent=4),
        )

        inverter_data = ModbusInverter(
            self._device_info[unit_key]["inverter"], inverter_raw)
        logger.debug(inverter_data)

        logger.info(
            LOGGING_DEVICE_INFO
            + ": {status}, AC {power_ac} W, DC {power_dc} W, {energytotal} kWh, "
            + "Grid status {grid_status}, ",
            unit_key=inverter_data.info.unit_key(":"),
            device="inverter",
            info=inverter_data.info,
            status=inverter_data.status,
            power_ac=inverter_data.ac.power.actual,
            power_dc=inverter_data.dc.power,
            energytotal=round(inverter_data.energytotal / 1000, 2),
            grid_status=inverter_data.grid_status,
        )

        return inverter_data

    def _map_meters(
        self, unit_key: str, meters_raw: dict[str, SunSpecPayload]
    ) -> dict[str, ModbusMeter]:
        meters: dict[str, ModbusMeter] = {}
        for meter_key, meter_raw in meters_raw.items():
            logger.debug(
                "Meter {meter} raw:\n{raw}",
                meter=meter_key,
                raw=json.dumps(meter_raw, indent=4),
            )

            meter_data = ModbusMeter(
                self._device_info[unit_key][meter_key], meter_raw)
            logger.debug(meter_data)
            logger.info(
                LOGGING_DEVICE_INFO +
                ": {power} W, {consumption} kWh, {delivery} kWh",
                unit_key=meter_data.info.unit_key(":"),
                device=meter_key,
                info=meter_data.info,
                power=meter_data.power.actual,
                consumption=round(meter_data.energy.totalimport / 1000, 3),
                delivery=round(meter_data.energy.totalexport / 1000, 3),
            )

            meters[meter_key] = meter_data

        return meters

    def _map_batteries(
        self, unit_key: str, batteries_raw: dict[str, SunSpecPayload]
    ) -> dict[str, ModbusBattery]:
        batteries = {}
        for battery_key, battery_raw in batteries_raw.items():
            logger.debug(
                "Battery {battery} raw:\n{raw}",
                battery=battery_key,
                raw=json.dumps(battery_raw, indent=4),
            )

            battery_data = ModbusBattery(
                self._device_info[unit_key][battery_key], battery_raw)
            logger.debug(battery_data)
            logger.info(
                LOGGING_DEVICE_INFO +
                ": {status}, {power} W, {state_of_charge} %",
                unit_key=battery_data.info.unit_key(":"),
                device=battery_key,
                info=battery_data.info,
                status=battery_data.status,
                power=battery_data.power,
                state_of_charge=battery_data.state_of_charge,
            )

            batteries[battery_key] = battery_data

        return batteries

    async def _handle_write_event(self, event: ModbusWriteEvent):
        await self._write_to_modbus(event.register, event.payload)

    async def _write_to_modbus(
        self, register: SunSpecRegister, value: SunSpecInputData
    ) -> None:
        logger.info(
            f"Writing {value} to register {register.address} ({register.name})"
        )

        value_decoded = register.encode_request(value)
        logger.trace(f"Encoded value: {value_decoded}")

        try:
            async with self.client:
                await self.client.write_registers(
                    register.address,
                    value_decoded,
                    device_id=self.settings.unit,
                )

        except ModbusException as error:
            logger.debug(f"Modbus write exception: {error}")
            logger.error(f"Unwriteable register {register.address}")



from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import ephem
from astral import LocationInfo
from astral.sun import azimuth, elevation, sun
from numpy import cos, pi, sin
from pandas import DataFrame, Series
from sklearn.base import BaseEstimator, TransformerMixin
from tzlocal import get_localzone_name

from solaredge2mqtt.core.logging import logger

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings.models import LocationSettings


LOCAL_TZ = get_localzone_name()


class BaseEncoder(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        self.features: list[str] | None = None
        self._feature_names_out: list[str] = []

    def fit(self, x_vector: DataFrame, *_) -> BaseEncoder:
        if not hasattr(x_vector, "columns"):
            raise AttributeError("x_vector has no columns")

        self.features = x_vector.columns.tolist()
        return self

    def _transform(self, x_vector: DataFrame) -> DataFrame:
        if self.features is None:
            raise AttributeError(f"Encoder {self.__class__} is not been fitted yet.")

        if not hasattr(x_vector, "columns"):
            raise AttributeError("x_vector has no columns")

        if not all(feature in self.features for feature in x_vector.columns):
            raise AttributeError(f"Columns {x_vector.columns} are not in the vector")

        if not all(feature in x_vector.columns for feature in self.features):
            raise AttributeError(f"Columns {self.features} are not in the vector")

        return x_vector

    def _save_feature_names_out(self, x_vector: DataFrame) -> DataFrame:
        self._feature_names_out = x_vector.columns.to_list()
        return x_vector

    def get_feature_names_out(self, *_) -> list[str]:
        return self._feature_names_out


class CategoricalEncoder(BaseEncoder):
    def transform(self, x_vector: DataFrame) -> DataFrame:
        x_vector = self._transform(x_vector).astype("category")
        return self._save_feature_names_out(x_vector)


class CyclicalEncoder(BaseEncoder):
    def __init__(self, **cycle_lengths: dict[str, int]) -> None:
        super().__init__()
        self.cycle_lengths: dict[str, int] = cycle_lengths

    def transform(self, x_vector: DataFrame) -> DataFrame:
        x_vector = self._transform(x_vector)
        for feature in self.features:
            cycle = self.cycle_lengths.get(feature, None)
            if not cycle:
                raise ValueError(f"Unknown cyclical feature {feature}")

            x_vector = self.transform_cycle_columns(
                x_vector, feature, x_vector[feature], cycle
            )
            x_vector.drop(feature, axis=1, inplace=True)

        return self._save_feature_names_out(x_vector)

    def get_params(self, deep=True) -> dict[str, int]:
        return self.cycle_lengths

    @staticmethod
    def transform_cycle_columns(
        x_vector: DataFrame, prefix: str, cycle_vector: DataFrame, cycle_length: float
    ) -> DataFrame:
        x_vector[f"{prefix}_cos"] = cos(2 * pi * cycle_vector / cycle_length)
        x_vector[f"{prefix}_sin"] = sin(2 * pi * cycle_vector / cycle_length)

        return x_vector


class TimeEncoder(BaseEncoder):
    def __init__(self) -> None:
        super().__init__()
        self.season_starts: dict[int, dict[str, datetime]] = {}

    def transform(self, x_vector: DataFrame) -> DataFrame:
        x_vector = self._transform(x_vector)
        for feature in self.features:
            x_vector = CyclicalEncoder.transform_cycle_columns(
                x_vector, f"{feature}_hour", x_vector[feature].dt.hour, 24
            )

            x_vector = CyclicalEncoder.transform_cycle_columns(
                x_vector, f"{feature}_month", x_vector[feature].dt.month, 12
            )

            x_vector[f"{feature}_dst"] = (
                x_vector[feature]
                .apply(lambda x: x.dst() != timedelta(0))
                .astype("category")
            )

            x_vector[f"{feature}_season"] = (
                x_vector[feature].apply(self._map_season).astype("category")
            )
            x_vector = CyclicalEncoder.transform_cycle_columns(
                x_vector,
                f"{feature}_day_of_year",
                x_vector[feature].dt.dayofyear,
                365.25,
            )

            x_vector.drop(feature, axis=1, inplace=True)

        logger.trace(x_vector.head(30))
        return self._save_feature_names_out(x_vector)

    def _map_season(self, date: datetime) -> str:
        year = date.year
        if year not in self.season_starts:
            equinox_mar = ephem.next_vernal_equinox(str(year))
            solstice_jun = ephem.next_summer_solstice(equinox_mar)
            equinox_sep = ephem.next_autumnal_equinox(solstice_jun)
            solstice_dec = ephem.next_winter_solstice(equinox_sep)

            self.season_starts[year] = {
                "spring": equinox_mar.datetime().astimezone(),
                "summer": solstice_jun.datetime().astimezone(),
                "autumn": equinox_sep.datetime().astimezone(),
                "winter": solstice_dec.datetime().astimezone(),
            }

        starts = self.season_starts[year]

        season = "winter"

        if date >= starts["spring"]:
            if date < starts["summer"]:
                season = "spring"
            elif date < starts["autumn"]:
                season = "summer"
            elif date < starts["winter"]:
                season = "autumn"

        return season


class SunEncoder(BaseEncoder):
    def __init__(self, location: LocationSettings) -> None:
        super().__init__()
        self.location = location
        self._location = LocationInfo(
            "name",
            "region",
            timezone=LOCAL_TZ,
            latitude=location.latitude,
            longitude=location.longitude,
        )

    def transform(self, x_vector: DataFrame) -> DataFrame:
        x_vector = self._transform(x_vector)

        for feature in self.features:
            time_key = f"{feature}_time"

            x_vector[time_key] = x_vector[feature].apply(
                lambda x: x + timedelta(minutes=30)
            )

            x_vector[f"{feature}_elevation"] = x_vector[time_key].apply(
                lambda x: elevation(self._location.observer, x)
            )

            x_vector = CyclicalEncoder.transform_cycle_columns(
                x_vector,
                f"{feature}_azimuth",
                x_vector[time_key].apply(lambda x: azimuth(self._location.observer, x)),
                360,
            )

            x_vector[
                [
                    f"{feature}_daylight",
                    f"{feature}_delta_sunrise",
                    f"{feature}_delta_sunset",
                ]
            ] = x_vector[time_key].apply(self.daylight_info)

            x_vector.drop([feature, time_key], axis=1, inplace=True)

        return self._save_feature_names_out(x_vector)

    def daylight_info(self, row_time: datetime) -> Series:
        s = sun(self._location.observer, row_time)
        daylight = (s["sunset"] - s["sunrise"]).total_seconds() / 3600

        delta_sunrise = (row_time - s["sunrise"]).total_seconds() / 3600
        delta_sunset = (s["sunset"] - row_time).total_seconds() / 3600

        return Series([daylight, delta_sunrise, delta_sunset])



import json
from asyncio import Queue, QueueFull

from aiomqtt import Client, Message, Will
from pydantic import BaseModel, ValidationError

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.models import BaseInputField
from solaredge2mqtt.core.mqtt.events import (
    MQTTPublishEvent,
    MQTTReceivedEvent,
    MQTTSubscribeEvent,
)
from solaredge2mqtt.core.mqtt.models import MAX_MQTT_PAYLOAD_SIZE
from solaredge2mqtt.core.mqtt.settings import MQTTSettings


class MQTTClient(Client):
    def __init__(self, settings: MQTTSettings, event_bus: EventBus):
        self.broker = settings.broker
        self.port = settings.port

        self.topic_prefix = settings.topic_prefix

        logger.info(
            "Using MQTT broker: {broker}:{port}",
            broker=settings.broker,
            port=settings.port,
        )

        will = Will(
            topic=f"{self.topic_prefix}/status", payload="offline", qos=1, retain=True
        )

        self._subscribed_topics: dict[str, type[BaseInputField]] = {}

        self._received_message_queue: Queue[Message] = Queue(maxsize=10)

        self.event_bus = event_bus
        self._subscribe_events()

        super().__init__(
            self.broker,
            self.port,
            will=will,
            **settings.kargs,
        )

    def _subscribe_events(self) -> None:
        self.event_bus.unsubscribe_all(MQTTPublishEvent)
        self.event_bus.unsubscribe_all(MQTTSubscribeEvent)

        self.event_bus.subscribe(MQTTPublishEvent, self.event_listener)
        self.event_bus.subscribe(MQTTSubscribeEvent, self._subscribe_topic)

    async def _subscribe_topic(self, event: MQTTSubscribeEvent) -> None:
        if event.topic not in self._subscribed_topics:
            logger.info(f"Subscribing to topic: {event.topic}")
            self._subscribed_topics[event.topic] = event.model
            await self.subscribe(event.topic)

    async def listen(self) -> None:
        if self._subscribed_topics:
            async for message in self.messages:
                topic = str(message.topic)
                if topic not in self._subscribed_topics:
                    logger.warning(
                        f"Received message on unsubscribed topic: {topic}"
                    )
                    continue
                if len(message.payload) > MAX_MQTT_PAYLOAD_SIZE:
                    logger.warning(
                        f"Payload too large on topic: {topic} "
                        f"({len(message.payload)} bytes)"
                    )
                    continue
                try:
                    self._received_message_queue.put_nowait(message)
                except QueueFull:
                    logger.warning(
                        "MQTT processing queue full – dropping message")

    async def process_queue(self) -> None:
        if self._subscribed_topics:
            while True:
                message = await self._received_message_queue.get()
                try:
                    await self._handle_message(message)
                except Exception as ex:
                    logger.error(f"Error while processing MQTT message: {ex}")

    async def _handle_message(self, message: Message) -> None:
        topic = str(message.topic)
        try:
            model = self._subscribed_topics.get(topic)
            if not model:
                logger.warning(
                    f"Received message for unexpected topic: {topic}")
                return

            payload = message.payload.decode()

            try:
                input_raw = json.loads(payload)
            except json.JSONDecodeError:
                input_raw = json.loads(json.dumps(payload))

            if isinstance(input_raw, dict):
                parsed_input = model(**input_raw)
            else:
                parsed_input = model(input_raw)

            await self.event_bus.emit(
                MQTTReceivedEvent(topic, parsed_input)
            )
        except (ValidationError, json.JSONDecodeError, TypeError) as ex:
            logger.warning(
                f"Received invalid message on topic: {topic}, error: {ex}"
            )

    async def publish_status_online(self) -> None:
        await self.publish_to("status", "online", True)

    async def publish_status_offline(self) -> None:
        await self.publish_to("status", "offline", True)

    async def event_listener(self, event: MQTTPublishEvent) -> None:
        await self.publish_to(
            event.topic,
            event.payload,
            event.retain,
            event.qos,
            event.topic_prefix,
            event.exclude_none,
        )

    async def publish_to(
        self,
        topic: str,
        payload: str | int | float | BaseModel,
        retain: bool,
        qos: int = 1,
        topic_prefix: str | None = None,
        exclude_none: bool = False,
    ) -> None:
        if self._connected:
            topic = f"{topic_prefix or self.topic_prefix}/{topic}"

            if isinstance(payload, BaseModel):
                payload = payload.model_dump_json(exclude_none=exclude_none)

            await self.publish(topic, payload, qos=qos, retain=retain)



from __future__ import annotations

from pydantic import Field

from solaredge2mqtt.core.models import Solaredge2MQTTBaseModel
from solaredge2mqtt.services.modbus.models.base import ModbusUnitInfo
from solaredge2mqtt.services.modbus.models.battery import ModbusBattery
from solaredge2mqtt.services.modbus.models.inverter import ModbusInverter
from solaredge2mqtt.services.modbus.models.meter import ModbusMeter


class ModbusUnit(Solaredge2MQTTBaseModel):
    info: ModbusUnitInfo | None
    inverter: ModbusInverter
    meters: dict[str, ModbusMeter] = Field(default_factory=dict)
    batteries: dict[str, ModbusBattery] = Field(default_factory=dict)

    def has_unit_info(self) -> bool:
        return self.info is not None



from __future__ import annotations

import time
from asyncio import Event, to_thread
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING

from joblib import Memory
from numpy import percentile
from pandas import DataFrame
from sklearn import clone
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.compose import ColumnTransformer
from sklearn.ensemble import HistGradientBoostingRegressor
from sklearn.inspection import permutation_importance
from sklearn.model_selection import GridSearchCV, TimeSeriesSplit, train_test_split
from sklearn.pipeline import Pipeline
from tzlocal import get_localzone_name

from solaredge2mqtt.core.events import EventBus
from solaredge2mqtt.core.exceptions import InvalidDataException
from solaredge2mqtt.core.influxdb import InfluxDBAsync, Point
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.mqtt.events import MQTTPublishEvent
from solaredge2mqtt.core.timer.events import Interval10MinTriggerEvent
from solaredge2mqtt.services.forecast.encoders import (
    CategoricalEncoder,
    CyclicalEncoder,
    SunEncoder,
    TimeEncoder,
)
from solaredge2mqtt.services.forecast.events import ForecastEvent
from solaredge2mqtt.services.forecast.models import Forecast, ForecasterType
from solaredge2mqtt.services.forecast.settings import ForecastSettings
from solaredge2mqtt.services.weather.events import WeatherUpdateEvent
from solaredge2mqtt.services.weather.models import OpenWeatherMapForecastData

if TYPE_CHECKING:
    from solaredge2mqtt.core.settings.models import LocationSettings


LOCAL_TZ = get_localzone_name()


class ForecastService:
    def __init__(
        self,
        settings: ForecastSettings,
        location: LocationSettings,
        event_bus: EventBus,
        influxdb: InfluxDBAsync,
    ) -> None:
        self.settings = settings
        self.location = location
        self.event_bus = event_bus
        self._subscribe_events()

        self.influxdb = influxdb

        self.forecasters: dict[ForecasterType, Forecaster] = {
            typed: Forecaster(typed, location, settings)
            for typed in ForecasterType
        }

        self.last_weather_forecast: list[OpenWeatherMapForecastData] | None = None
        self.last_hour_forecast: dict[int,
                                      OpenWeatherMapForecastData] | None = None

    def _subscribe_events(self) -> None:
        self.event_bus.subscribe(WeatherUpdateEvent, self.weather_update)
        self.event_bus.subscribe(Interval10MinTriggerEvent, self.forecast_loop)

    async def weather_update(self, event: WeatherUpdateEvent) -> None:
        self.last_weather_forecast = event.weather.hourly

        if self.last_hour_forecast is None:
            self.last_hour_forecast = {}

        self.last_hour_forecast[event.weather.hourly[0].hour] = event.weather.hourly[0]

        now = datetime.now().astimezone()
        last_hour = now - timedelta(hours=1)

        for delete_hour in range(0, 24):
            if delete_hour in [now.hour, last_hour.hour]:
                continue

            self.last_hour_forecast.pop(delete_hour, None)

        logger.debug(self.last_hour_forecast)

        if last_hour.hour in self.last_hour_forecast:
            await self.write_new_training_data(self.last_hour_forecast[last_hour.hour])

    async def write_new_training_data(
        self, last_hour_weather_forecast: OpenWeatherMapForecastData
    ) -> None:
        now = datetime.now().astimezone()
        last_hour = now.replace(
            minute=0, second=0, microsecond=0) - timedelta(hours=1)

        training_data = last_hour_weather_forecast.model_dump_estimation_data()
        training_data["time"] = last_hour
        training_data = await self.add_last_hour_pv_production(training_data)
        await self.write_new_training_data_to_influxdb(training_data)

        if (now.minute // 10) * 10 == 20:
            await self.train()

    async def add_last_hour_pv_production(
        self, trainings_data
    ) -> dict[str, str | float | int | None]:
        production_data = await self.influxdb.query_first("production")

        if production_data is None:
            raise InvalidDataException(
                "Missing production data of last hour for forecast training."
                + " Did the service write power information to InfluxDB?"
            )

        trainings_data[ForecasterType.POWER.target_column] = round(
            production_data[ForecasterType.POWER.target_column]
        )
        trainings_data[ForecasterType.ENERGY.target_column] = round(
            production_data[ForecasterType.ENERGY.target_column], 2
        )
        return trainings_data

    async def write_new_training_data_to_influxdb(self, trainings_data):
        point = Point("forecast_training")
        for key, value in trainings_data.items():
            if isinstance(value, (int, float, str, bool)):
                point.field(key, value)

        point.time(trainings_data["time"].astimezone(timezone.utc))

        logger.info("Write new forecast training data to influxdb")
        logger.debug(trainings_data)
        await self.influxdb.write_point(point)

    async def train(self) -> None:
        data = await self.influxdb.query_dataframe("training_data")
        data["time"] = data["_time"].dt.tz_convert(LOCAL_TZ)
        await to_thread(self.training, data)

    def training(self, data: DataFrame) -> None:
        for forecaster in self.forecasters.values():
            forecaster.train(data)

    async def forecast_loop(self, _):
        if (
            not self.forecasters[ForecasterType.ENERGY].is_trained
            or not self.forecasters[ForecasterType.POWER].is_trained
        ):
            await self.train()

        if self.last_weather_forecast is None:
            raise InvalidDataException(
                "Missing weather forecast for production forecast"
            )

        estimation_data_list = [
            {
                "time": datetime(
                    year=weather_forecast.year,
                    month=weather_forecast.month,
                    day=weather_forecast.day,
                    hour=weather_forecast.hour,
                    minute=0,
                    second=0,
                    microsecond=0,
                ).astimezone(),
                **weather_forecast.model_dump_estimation_data(),
            }
            for weather_forecast in self.last_weather_forecast
        ]

        data = DataFrame(estimation_data_list)
        data["time"] = data["time"].astype(f"datetime64[ns, {LOCAL_TZ}]")

        for typed, forecaster in self.forecasters.items():
            predicted_data = await forecaster.predict(data)
            await self._write_periods_to_influxdb(predicted_data, typed)

        await self.publish_forecast()

    async def _write_periods_to_influxdb(
        self, periods: DataFrame, typed: ForecasterType
    ) -> None:
        points = []
        for _, period in periods.iterrows():
            point = Point("forecast")
            point.field(typed.target_column, period[typed.target_column])
            utc_time = period["time"].astimezone(timezone.utc)
            point.time(utc_time)
            points.append(point)

        logger.info("Write forecast data to influxdb")
        await self.influxdb.write_points(points)

    async def publish_forecast(self) -> None:
        forecast_data = await self.influxdb.query_dataframe("forecast")
        if not forecast_data.empty:
            forecast_data["time"] = forecast_data["_time"].dt.tz_convert(
                LOCAL_TZ)
            power_hours = {
                row["time"]: row["power"] for idx, row in forecast_data.iterrows()
            }
            energy_hours = {
                row["time"]: round(row["energy"] * 1000)
                for idx, row in forecast_data.iterrows()
            }
            forecast = Forecast(power_period=power_hours,
                                energy_period=energy_hours)
            logger.debug(forecast)

            await self.event_bus.emit(
                MQTTPublishEvent(
                    forecast.mqtt_topic(),
                    forecast,
                    self.settings.retain,
                )
            )
            await self.event_bus.emit(ForecastEvent(forecast))


class Forecaster:
    NUMERIC_FEATURES: list[str] = [
        "clouds",
        "dew_point",
        "feels_like",
        "humidity",
        "pop",
        "pressure",
        "rain",
        "temp",
        "uvi",
        "visibility",
        "wind_speed",
        "wind_gust",
    ]
    CATEGORICAL_FEATURES: list[str] = ["weather_id", "weather_main"]
    CYCLICAL_FEATURES: dict[str, int] = {
        "wind_deg": 360,
    }
    TARGET_FEATURES: dict[str, str] = {
        ForecasterType.ENERGY.target_column: "energy",
        ForecasterType.POWER.target_column: "power",
    }

    def __init__(
        self,
        typed: ForecasterType,
        location: LocationSettings,
        settings: ForecastSettings
    ) -> None:
        self.typed: ForecasterType = typed
        self.location = location
        self.enable_hyperparameter_tuning = settings.hyperparametertuning
        self.model_pipeline: Pipeline | None = None
        self.training_completed: Event = Event()

        self.memory = (
            Memory(settings.cachingdir, verbose=0)
            if settings.is_caching_enabled
            else False
        )

    def train(self, data: DataFrame) -> None:
        data_count = len(data)
        logger.info(
            f"Training model {self.typed} with {data_count} hours of data points"
        )

        if data_count < 60:
            raise InvalidDataException(
                "Forecast needs at least 60 hours of data at least to start training",
            )

        self.training_completed.clear()
        start_time = time.time()
        y_vector = data[self.typed.target_column]

        pipeline = self._prepare_model_pipeline(data.columns.to_list())

        if self.enable_hyperparameter_tuning:
            self.model_pipeline = self._hyperparametertuning(
                data, y_vector, pipeline)
        else:
            self.model_pipeline = pipeline

        self.model_pipeline.fit(data, y_vector)

        execution_time = time.time() - start_time
        self.training_completed.set()

        transformed_features = self.model_pipeline[
            "preprocessor"
        ].get_feature_names_out()
        logger.debug(
            "Transformed features ({count}): {features} ",
            count=len(transformed_features),
            features=", ".join(transformed_features),
        )

        selected_features = self.model_pipeline[
            "feature_selector"
        ].important_features_.tolist()
        logger.info(
            "Selected features ({count}): {features} ",
            count=len(selected_features),
            features=", ".join(selected_features),
        )

        logger.info(f"Training execution time: {execution_time:.2f} seconds")

    def _hyperparametertuning(self, data, y_vector, pipeline):
        param_grid = {
            "model__max_iter": [100, 200, 300],
            "model__max_depth": [None, 5, 10],
            "model__learning_rate": [0.01, 0.1],
        }

        grid_search = GridSearchCV(
            estimator=pipeline,
            param_grid=param_grid,
            cv=TimeSeriesSplit(n_splits=2),
            scoring="neg_mean_squared_error",
        )
        grid_search.fit(data, y_vector)

        logger.info(
            "Training with best parameters: {params}", params=grid_search.best_params_
        )
        logger.info(
            "Training with best score: {score}", score=grid_search.best_score_)

        return clone(grid_search.best_estimator_)

    def _prepare_model_pipeline(self, x_vector_columns: list[str]) -> Pipeline:
        base_estimator = HistGradientBoostingRegressor(
            random_state=42,
            categorical_features="from_dtype",
            learning_rate=0.1,
        )

        return Pipeline(
            steps=[
                ("preprocessor", self._prepare_preprocessor(x_vector_columns)),
                (
                    "feature_selector",
                    PFISelector(
                        estimator=clone(base_estimator)
                    ),
                ),
                (
                    "model", clone(base_estimator)
                ),
            ],
            memory=self.memory,
        )

    def _prepare_preprocessor(self, x_vector_columns: list[str]) -> ColumnTransformer:
        ct = ColumnTransformer(
            transformers=[
                (
                    "cyc",
                    CyclicalEncoder(**self.CYCLICAL_FEATURES),
                    self._extract_used_columns(
                        self.CYCLICAL_FEATURES, x_vector_columns
                    ),
                ),
                (
                    "num",
                    "passthrough",
                    self._extract_used_columns(
                        self.NUMERIC_FEATURES, x_vector_columns),
                ),
                (
                    "time",
                    TimeEncoder(),
                    ["time"],
                ),
                ("sun", SunEncoder(self.location), ["time"]),
                (
                    "cat",
                    CategoricalEncoder(),
                    self._extract_used_columns(
                        self.CATEGORICAL_FEATURES, x_vector_columns
                    ),
                ),
            ],
            remainder="drop",
        )
        ct.set_output(transform="pandas")
        return ct

    @staticmethod
    def _extract_used_columns(
        typed_features: list[str] | dict[str, int], x_vector_columns: list[str]
    ) -> list[str]:
        return [col for col in typed_features if col in x_vector_columns]

    @property
    def is_trained(self) -> bool:
        return self.model_pipeline is not None

    async def predict(self, new_data: DataFrame) -> DataFrame:
        if self.model_pipeline is None:
            raise InvalidDataException("The model has not been trained yet.")

        data_for_prediction = new_data.copy()

        await self.training_completed.wait()

        predictions = self.model_pipeline.predict(data_for_prediction)
        data_for_prediction[self.typed.target_column] = predictions
        data_for_prediction[self.typed.target_column] = data_for_prediction[
            self.typed.target_column
        ].apply(self.typed.prepare_value)

        return data_for_prediction


class PFISelector(BaseEstimator, TransformerMixin):
    def __init__(self, estimator, n_repeats=10):
        self.estimator = estimator
        self.n_repeats = n_repeats

    def fit(self, x_vector: DataFrame, y_vector=None) -> PFISelector:
        x_train, x_test, y_train, y_test = train_test_split(
            x_vector, y_vector, test_size=0.1, random_state=42
        )
        self.estimator_ = self.estimator.fit(x_train, y_train)
        results = permutation_importance(
            self.estimator_,
            x_test,
            y_test,
            n_repeats=self.n_repeats,
            random_state=42,
            n_jobs=-1,
        )

        self.feature_importances_ = results.importances_mean

        threshold_value = percentile(self.feature_importances_, 75)

        self.important_indices_ = self.feature_importances_ > threshold_value
        self.important_features_ = x_vector.columns[self.important_indices_]
        return self

    def transform(self, x_vector: DataFrame) -> DataFrame:
        if not hasattr(self, "important_features_"):
            raise RuntimeError("PFISelector has not been fitted yet.")
        return x_vector[self.important_features_]

    def get_support(self, *_) -> list[bool]:
        if not hasattr(self, "important_indices_"):
            raise RuntimeError("PFISelector has not been fitted yet.")
        return self.important_indices_.to_list()



from solaredge2mqtt.services.modbus.sunspec.base import SunSpecOffset, SunSpecRegister
from solaredge2mqtt.services.modbus.sunspec.values import SunSpecValueType


class SunSpecMeterInfoRegister(SunSpecRegister):
    C_MANUFACTURER = "c_manufacturer", 40123, SunSpecValueType.STRING, True, 16
    C_MODEL = "c_model", 40139, SunSpecValueType.STRING, True, 16
    C_OPTION = "c_option", 40155, SunSpecValueType.STRING, True, 8
    C_VERSION = "c_version", 40163, SunSpecValueType.STRING, True, 8
    C_SERIALNUMBER = "c_serialnumber", 40171, SunSpecValueType.STRING, True, 16
    C_DEVICEADDRESS = "c_deviceaddress", 40187, SunSpecValueType.UINT16, True
    C_SUNSPEC_DID = "c_sunspec_did", 40188, SunSpecValueType.UINT16, True
    C_SUNSPEC_LENGTH = (
        "c_sunspec_length",
        40189,
        SunSpecValueType.UINT16,
    )


class SunSpecMeterRegister(SunSpecRegister):
    CURRENT = "current", 40190, SunSpecValueType.INT16, True
    L1_CURRENT = "l1_current", 40191, SunSpecValueType.INT16
    L2_CURRENT = "l2_current", 40192, SunSpecValueType.INT16
    L3_CURRENT = "l3_current", 40193, SunSpecValueType.INT16
    CURRENT_SCALE = "current_scale", 40194, SunSpecValueType.INT16, True

    VOLTAGE_LN = "voltage_ln", 40195, SunSpecValueType.INT16, True
    L1N_VOLTAGE = "l1n_voltage", 40196, SunSpecValueType.INT16
    L2N_VOLTAGE = "l2n_voltage", 40197, SunSpecValueType.INT16
    L3N_VOLTAGE = "l3n_voltage", 40198, SunSpecValueType.INT16
    VOLTAGE_LL = "voltage_lln", 40199, SunSpecValueType.INT16, True
    L12_VOLTAGE = "l12_voltage", 40200, SunSpecValueType.INT16
    L23_VOLTAGE = "l23_voltage", 40201, SunSpecValueType.INT16
    L31_VOLTAGE = "l31_voltage", 40202, SunSpecValueType.INT16
    VOLTAGE_SCALE = "voltage_scale", 40203, SunSpecValueType.INT16, True

    FREQUENCY = "frequency", 40204, SunSpecValueType.UINT16, True
    FREQUENCY_SCALE = "frequency_scale", 40205, SunSpecValueType.INT16, True

    POWER = "power", 40206, SunSpecValueType.INT16, True
    L1_POWER = "l1_power", 40207, SunSpecValueType.INT16
    L2_POWER = "l2_power", 40208, SunSpecValueType.INT16
    L3_POWER = "l3_power", 40209, SunSpecValueType.INT16
    POWER_SCALE = "power_scale", 40210, SunSpecValueType.INT16, True

    POWER_APPARENT = "power_apparent", 40211, SunSpecValueType.INT16, True
    L1_POWER_APPARENT = "l1_power_apparent", 40212, SunSpecValueType.INT16
    L2_POWER_APPARENT = "l2_power_apparent", 40213, SunSpecValueType.INT16
    L3_POWER_APPARENT = "l3_power_apparent", 40214, SunSpecValueType.INT16
    POWER_APPARENT_SCALE = "power_apparent_scale", 40215, SunSpecValueType.INT16, True

    POWER_REACTIVE = "power_reactive", 40216, SunSpecValueType.INT16, True
    L1_POWER_REACTIVE = "l1_power_reactive", 40217, SunSpecValueType.INT16
    L2_POWER_REACTIVE = "l2_power_reactive", 40218, SunSpecValueType.INT16
    L3_POWER_REACTIVE = "l3_power_reactive", 40219, SunSpecValueType.INT16
    POWER_REACTIVE_SCALE = "power_reactive_scale", 40220, SunSpecValueType.INT16, True

    POWER_FACTOR = "power_factor", 40221, SunSpecValueType.INT16, True
    L1_POWER_FACTOR = "l1_power_factor", 40222, SunSpecValueType.INT16
    L2_POWER_FACTOR = "l2_power_factor", 40223, SunSpecValueType.INT16
    L3_POWER_FACTOR = "l3_power_factor", 40224, SunSpecValueType.INT16
    POWER_FACTOR_SCALE = "power_factor_scale", 40225, SunSpecValueType.INT16, True

    EXPORT_ENERGY_ACTIVE = "export_energy_active", 40226, SunSpecValueType.UINT32, True
    L1_EXPORT_ENERGY_ACTIVE = (
        "l1_export_energy_active",
        40228,
        SunSpecValueType.UINT32,
    )
    L2_EXPORT_ENERGY_ACTIVE = (
        "l2_export_energy_active",
        40230,
        SunSpecValueType.UINT32,
    )
    L3_EXPORT_ENERY_ACTIVE = (
        "l3_export_energy_active",
        40232,
        SunSpecValueType.UINT32,
    )
    IMPORT_ENERGY_ACTIVE = "import_energy_active", 40234, SunSpecValueType.UINT32, True
    L1_IMPORT_ENERGY_ACTIVE = (
        "l1_import_energy_active",
        40236,
        SunSpecValueType.UINT32,
    )
    L2_IMPORT_ENERGY_ACTIVE = (
        "l2_import_energy_active",
        40238,
        SunSpecValueType.UINT32,
    )
    L3_IMPORT_ENERGY_ACTIVE = (
        "l3_import_energy_active",
        40240,
        SunSpecValueType.UINT32,
    )
    ENERGY_ACTIVE_SCALE = "energy_active_scale", 40242, SunSpecValueType.INT16, True

    EXPORT_ENERGY_APPARENT = "export_energy_apparent", 40243, SunSpecValueType.UINT32
    L1_EXPORT_ENERGY_APPARENT = (
        "l1_export_energy_apparent",
        40245,
        SunSpecValueType.UINT32,
    )
    L2_EXPORT_ENERGY_APPARENT = (
        "l2_export_energy_apparent",
        40247,
        SunSpecValueType.UINT32,
    )
    L3_EXPORT_ENERGY_APPARENT = (
        "l3_export_energy_apparent",
        40249,
        SunSpecValueType.UINT32,
    )
    IMPORT_ENERGY_APPARENT = "import_energy_apparent", 40251, SunSpecValueType.UINT32
    L1_IMPORT_ENERGY_APPARENT = (
        "l1_import_energy_apparent",
        40253,
        SunSpecValueType.UINT32,
    )
    L2_IMPORT_ENERGY_APPARENT = (
        "l2_import_energy_apparent",
        40255,
        SunSpecValueType.UINT32,
    )
    L3_IMPORT_ENERGY_APPARENT = (
        "l3_import_energy_apparent",
        40257,
        SunSpecValueType.UINT32,
    )
    ENERGY_APPARENT_SCALE = "energy_apparent_scale", 40259, SunSpecValueType.INT16

    IMPORT_ENERGY_REACTIVE_Q1 = (
        "import_energy_reactive_q1",
        40260,
        SunSpecValueType.UINT32,
    )
    L1_IMPORT_ENERGY_REACTIVE_Q1 = (
        "l1_import_energy_reactive_q1",
        40262,
        SunSpecValueType.UINT32,
    )
    L2_IMPORT_ENERGY_REACTIVE_Q1 = (
        "l2_import_energy_reactive_q1",
        40264,
        SunSpecValueType.UINT32,
    )
    L3_IMPORT_ENERGY_REACTIVE_Q1 = (
        "l3_import_energy_reactive_q1",
        40266,
        SunSpecValueType.UINT32,
    )
    IMPORT_ENERGY_REACTIVE_Q2 = (
        "import_energy_reactive_q2",
        40268,
        SunSpecValueType.UINT32,
    )
    L1_IMPORT_ENERGY_REACTIVE_Q2 = (
        "l1_import_energy_reactive_q2",
        40270,
        SunSpecValueType.UINT32,
    )
    L2_IMPORT_ENERGY_REACTIVE_Q2 = (
        "l2_import_energy_reactive_q2",
        40272,
        SunSpecValueType.UINT32,
    )
    L3_IMPORT_ENERGY_REACTIVE_Q2 = (
        "l3_import_energy_reactive_q2",
        40274,
        SunSpecValueType.UINT32,
    )
    EXPORT_ENERGY_REACTIVE_Q3 = (
        "export_energy_reactive_q3",
        40276,
        SunSpecValueType.UINT32,
    )
    L1_EXPORT_ENERGY_REACTIVE_Q3 = (
        "l1_export_energy_reactive_q3",
        40278,
        SunSpecValueType.UINT32,
    )
    L2_EXPORT_ENERGY_REACTIVE_Q3 = (
        "l2_export_energy_reactive_q3",
        40280,
        SunSpecValueType.UINT32,
    )
    L3_EXPORT_ENERGY_REACTIVE_Q3 = (
        "l3_export_energy_reactive_q3",
        40282,
        SunSpecValueType.UINT32,
    )
    EXPORT_ENERGY_REACTIVE_Q4 = (
        "export_energy_reactive_q4",
        40284,
        SunSpecValueType.UINT32,
    )
    L1_EXPORT_ENERGY_REACTIVE_Q4 = (
        "l1_export_energy_reactive_q4",
        40286,
        SunSpecValueType.UINT32,
    )
    L2_EXPORT_ENERGY_REACTIVE_Q4 = (
        "l2_export_energy_reactive_q4",
        40288,
        SunSpecValueType.UINT32,
    )
    L3_EXPORT_ENERGY_REACTIVE_Q4 = (
        "l3_export_energy_reactive_q4",
        40290,
        SunSpecValueType.UINT32,
    )
    ENERGY_REACTIVE_SCALE = "energy_reactive_scale", 40292, SunSpecValueType.INT16


class SunSpecMeterOffset(SunSpecOffset):
    METER0 = "meter0", 0
    METER1 = "meter1", 174
    METER2 = "meter2", 348



import json
from importlib.metadata import version as pkg_version
from pathlib import Path

try:
    from setuptools_scm import get_version
except ImportError:
    get_version = None

__version__ = "0.0.0-unknown"

version_file = Path(__file__).parent / "version.json"
if version_file.exists():
    try:
        with version_file.open(encoding="utf-8") as f:
            __version__ = json.load(f)["version"]
    except Exception:
        pass
elif get_version is not None:
    try:
        __version__ = get_version(root="..", relative_to=__file__)
    except Exception:
        pass
else:
    try:
        __version__ = pkg_version("solaredge2mqtt")
    except Exception:
        pass



from pydantic import Field, field_serializer

from solaredge2mqtt.core.models import Solaredge2MQTTBaseModel
from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantBinarySensorType as HABinarySensor,
)
from solaredge2mqtt.services.homeassistant.models import (
    HomeAssistantSensorType as HASensor,
)
from solaredge2mqtt.services.models import Component


class WallboxInfo(Solaredge2MQTTBaseModel):
    manufacturer: str
    model: str
    version: str
    serialnumber: str

    def __init__(self, data: dict[str, str | int]) -> dict[str, str]:
        values = {
            "manufacturer": "SolarEdge",
            "model": data["model"],
            "version": data["firmwareVersion"],
            "serialnumber": data["serialNumber"],
        }

        super().__init__(**values)

    def homeassistant_device_info(self) -> dict[str, any]:
        return {
            "name": f"{self.manufacturer} Wallbox",
            "manufacturer": self.manufacturer,
            "model": self.model,
            "hw_version": self.version,
            "serial_number": self.serialnumber,
        }


class WallboxAPI(Component):
    COMPONENT = "wallbox"
    SOURCE = "api"

    info: WallboxInfo
    power: float = Field(**HASensor.POWER_W.field("Power"))
    state: str = Field(title="State")
    vehicle_plugged: bool = Field(
        **HABinarySensor.PLUG.field("Vehicle plugged"))
    max_current: float = Field(**HASensor.CURRENT_A.field("Max current"))

    def __init__(self, data: dict[str, str | int]):
        info = WallboxInfo(data)
        power = round(data["meter"]["totalActivePower"] / 1000)
        state = data["state"]
        vehicle_connected = bool(data["vehiclePlugged"])
        max_current = float(data["maxCurrent"])

        super().__init__(
            info=info,
            power=power,
            state=state,
            vehicle_plugged=vehicle_connected,
            max_current=max_current,
        )

    def homeassistant_device_info(self) -> dict[str, any]:
        return self.info.homeassistant_device_info()

    @classmethod
    # pylint: disable=arguments-differ
    def model_json_schema(cls, mode: str = "serialization") -> dict[str, any]:
        schema = super().model_json_schema(mode=mode)
        schema["properties"].pop("info", None)
        return schema

    @field_serializer("vehicle_plugged")
    def serialize_vehicle_plugged(self, value: bool) -> str:
        return "true" if value else "false"



from os import environ, listdir, path
from typing import Generator

from pydantic import BaseModel, Field

from solaredge2mqtt.core.influxdb.settings import InfluxDBSettings
from solaredge2mqtt.core.logging import logger
from solaredge2mqtt.core.logging.models import LoggingLevelEnum
from solaredge2mqtt.core.mqtt.settings import MQTTSettings
from solaredge2mqtt.services.energy.settings import EnergySettings, PriceSettings
from solaredge2mqtt.services.forecast.settings import ForecastSettings
from solaredge2mqtt.services.homeassistant.settings import HomeAssistantSettings
from solaredge2mqtt.services.modbus.settings import ModbusSettings
from solaredge2mqtt.services.monitoring.settings import MonitoringSettings
from solaredge2mqtt.services.powerflow.settings import PowerflowSettings
from solaredge2mqtt.services.wallbox.settings import WallboxSettings
from solaredge2mqtt.services.weather.settings import WeatherSettings

DOCKER_SECRETS_DIR = "/run/secrets"


class LocationSettings(BaseModel):
    latitude: float
    longitude: float


class ServiceSettings(BaseModel):
    interval: int = Field(5)
    logging_level: LoggingLevelEnum = LoggingLevelEnum.INFO

    modbus: ModbusSettings
    mqtt: MQTTSettings

    powerflow: PowerflowSettings = PowerflowSettings()
    energy: EnergySettings = EnergySettings()

    location: LocationSettings | None = None
    prices: PriceSettings = PriceSettings()

    monitoring: MonitoringSettings | None = None
    wallbox: WallboxSettings | None = None

    influxdb: InfluxDBSettings | None = None

    weather: WeatherSettings | None = None

    forecast: ForecastSettings | None = None

    homeassistant: HomeAssistantSettings | None = None

    def __init__(self, **data: dict[str, any]):
        sources = [self._read_environment,
                   self._read_dotenv, self._read_secrets]
        data = self._parse_key_and_values(sources, data)
        super().__init__(**data)

    @property
    def is_location_configured(self) -> bool:
        return self.location is not None

    @property
    def is_prices_configured(self) -> bool:
        return self.prices is not None and self.prices.is_configured

    @property
    def is_monitoring_configured(self) -> bool:
        return self.monitoring is not None and self.monitoring.is_configured

    @property
    def is_wallbox_configured(self) -> bool:
        return self.wallbox is not None and self.wallbox.is_configured

    @property
    def is_influxdb_configured(self) -> bool:
        return self.influxdb is not None and self.influxdb.is_configured

    @property
    def is_weather_configured(self) -> bool:
        is_configured = self.weather is not None and self.weather.is_configured

        if is_configured and not self.is_location_configured:
            logger.warning(
                "Weather settings are configured but location is not configured."
            )
            is_configured = False

        return is_configured

    @property
    def is_forecast_configured(self) -> bool:
        is_configured = self.forecast is not None and self.forecast.is_configured

        if is_configured and not self.is_location_configured:
            logger.warning(
                "Forecast settings are configured but location is not configured."
            )
            is_configured = False

        if is_configured and not self.is_weather_configured:
            logger.warning(
                "Forecast settings are configured but weather is not configured."
            )
            is_configured = False

        return is_configured

    @property
    def is_homeassistant_configured(self) -> bool:
        return self.homeassistant is not None and self.homeassistant.is_configured

    def _parse_key_and_values(
        self, sources: list[callable], data: dict[str, any]
    ) -> dict[str, any]:
        for source in sources:
            for key, value in source():
                key = key.lower().strip()[8:]
                subkeys = key.split("__")
                self.insert_nested_key(
                    data, subkeys, value.strip())

        return data

    @classmethod
    def insert_nested_key(cls, container: dict, keys: list[str], value: any) -> None:
        key, i = cls._identify_key_and_position(keys)

        key, idx, next_container = cls._get_or_initialize_nested_container(
            container, key, i
        )

        cls._insert_value_in_container(
            container, keys, value, key, idx, next_container)

    @classmethod
    def _identify_key_and_position(cls, keys: list[str]) -> tuple[str, int]:
        key = keys[0]
        for i in range(len(key) - 1, -1, -1):
            if not key[i].isdigit():
                break
        return key, i

    @classmethod
    def _get_or_initialize_nested_container(
        cls, container: dict, key: str, i: int
    ) -> tuple[str, int | str, dict | list]:
        prefix, idx = key[:i + 1], key[i + 1:]
        if idx.isdigit():
            key, idx = prefix, int(idx)
            if key not in container or not isinstance(container[key], list):
                container[key] = []
            while len(container[key]) <= idx:
                container[key].append({})
            next_container = container[key][idx]
        else:
            if key not in container or not isinstance(container[key], dict):
                container[key] = {}
            next_container = container[key]
        return key, idx, next_container

    @classmethod
    def _insert_value_in_container(
        cls,
        container: dict[str, any],
        keys: list[str],
        value: any,
        key: str,
        idx: int | str,
        next_container: dict | list,
    ) -> None:
        if len(keys) == 1:
            if isinstance(next_container, dict):
                if isinstance(container[key], list):
                    container[key][idx] = value
                else:
                    container[key] = value
            else:
                container[key] = value
        else:
            cls.insert_nested_key(next_container, keys[1:], value)

    @classmethod
    def _read_environment(cls) -> Generator[tuple[str, str], any, any]:
        for key, value in environ.items():
            if cls._has_prefix(key):
                yield key, value

    @classmethod
    def _read_secrets(cls) -> Generator[tuple[str, str], any, any]:
        if path.exists(DOCKER_SECRETS_DIR) and path.isdir(DOCKER_SECRETS_DIR):
            for filename in listdir(DOCKER_SECRETS_DIR):
                if cls._has_prefix(filename):
                    with open(
                        path.join(DOCKER_SECRETS_DIR, filename), "r", encoding="utf-8"
                    ) as f:
                        yield filename, f.read()

    @classmethod
    def _read_dotenv(cls) -> Generator[tuple[str, str], any, any]:
        if path.exists(".env"):
            with open(".env", "r", encoding="utf-8") as f:
                for line in f.readlines():
                    line = line.strip()
                    if cls._has_prefix(line) and "=" in line:
                        key, value = line.split("=", 1)
                        yield key, value

    @staticmethod
    def _has_prefix(key: str) -> bool:
        return key.lower().startswith("se2mqtt_")



from typing import ClassVar

from solaredge2mqtt.core.models import Solaredge2MQTTBaseModel


class ComponentValueGroup(Solaredge2MQTTBaseModel):
    @staticmethod
    def scale_value(
        data: dict[str, str | int],
        value_key: str,
        scale_key: str | None = None,
        digits: int = 2,
    ) -> float:
        if scale_key is None:
            scale_key = f"{value_key}_scale"

        value = int(data[value_key])
        scale = int(data[scale_key])

        return round(value * 10**scale, digits)


class Component(ComponentValueGroup):
    COMPONENT: ClassVar[str] = "unknown"
    SOURCE: ClassVar[str | None] = None

    @property
    def influxdb_tags(self) -> dict[str, str]:
        return {
            "component": self.COMPONENT,
            "source": self.SOURCE,
        }

    def mqtt_topic(self) -> str:
        if self.SOURCE:
            topic = f"{self.SOURCE}/{self.COMPONENT}"
        else:
            topic = self.COMPONENT

        return topic

    def __str__(self) -> str:
        if self.SOURCE:
            name = f"{self.SOURCE}: {self.COMPONENT}"
        else:
            name = self.COMPONENT

        return name



import asyncio
from typing import Callable

import aiohttp

from solaredge2mqtt.core.logging import logger


class HTTPClientAsync:
    def __init__(self, service: str) -> None:
        self.service = service
        self.session: aiohttp.ClientSession | None = None

    def init(self) -> None:
        if not self.session:
            cookie_jar = aiohttp.CookieJar(unsafe=True)
            self.session = aiohttp.ClientSession(cookie_jar=cookie_jar)

    async def _handle_response(
        self, response: aiohttp.ClientResponse, expect_json: bool = True
    ) -> dict | str | None:
        try:
            response.raise_for_status()
            return await response.json() if expect_json else await response.text()
        except aiohttp.ClientResponseError as error:
            logger.warning(f"HTTP error from {self.service}: {error}")
            raise error

    async def _get(
        self,
        url: str,
        params: dict[str, str | int | float] | None = None,
        headers: dict[str, str] | None = None,
        verify: bool = True,
        login: Callable | None = None,
    ) -> dict | None:
        try:
            self.init()

            async with self.session.get(
                url, params=params, headers=headers, ssl=verify
            ) as response:
                if response.status in (401, 403) and login:
                    await login()
                    return await self._get(url, params, headers, verify, None)
                return await self._handle_response(response)
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError) as error:
            logger.warning(f"Connection issue with {self.service}: {error}")
            return None

    async def _post(
        self,
        url: str,
        json: dict[str, str] | None = None,
        data: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        verify: bool = True,
        expect_json: bool = True,
        login: Callable | None = None,
    ) -> dict | str | None:
        try:
            self.init()

            async with self.session.post(
                url, json=json, data=data, headers=headers, ssl=verify
            ) as response:
                if response.status in (401, 403) and login:
                    await login()
                    return await self._post(
                        url, json, data, headers, verify, expect_json, None
                    )

                return await self._handle_response(response, expect_json)
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError) as error:
            logger.warning(f"Connection issue with {self.service}: {error}")
            return None

    async def close(self) -> None:
        if self.session:
            await self.session.close()
            self.session = None

    def cookie_exists(self, cookie_name: str) -> bool:
        return self.get_cookie(cookie_name) is not None

    def get_cookie(self, cookie_name: str) -> str | None:
        value: str | None = None
        if self.session:
            for cookie in self.session.cookie_jar:
                if cookie.key == cookie_name:
                    value = cookie.value
                    break

        return value



from typing import Type

from pymodbus.client import ModbusTcpClient

from solaredge2mqtt.core.models import EnumModel

SunSpecRawData = str | int | float

SunSpecInputData = int | float | str | bool

SunSpecPayload = dict[str, SunSpecRawData]


C_SUNSPEC_DID_MAP = {
    101: "Single Phase Inverter",
    102: "Split Phase Inverter",
    103: "Three Phase Inverter",
    201: "Single Phase Meter",
    202: "Split Phase Meter",
    203: "Wye 3P1N Three Phase Meter",
    204: "Delta 3P Three Phase Meter",
    802: "Battery",
    803: "Lithium Ion Bank Battery",
    804: "Lithium Ion String Battery",
    805: "Lithium Ion Module Battery",
    806: "Flow Battery",
    807: "Flow String Battery",
    808: "Flow Module Battery",
    809: "Flow Stack Battery",
}

INVERTER_STATUS_MAP = {
    1: "Off",
    2: "Sleeping (auto-shutdown) – Night mode",
    3: "Grid Monitoring/wake-up",
    4: "Inverter is ON and producing power",
    5: "Production (curtailed)",
    6: "Shutting down",
    7: "Fault",
    8: "Maintenance/setup",
}

BATTERY_STATUS_MAP = {
    0: "Off",
    1: "Standby",
    2: "Initializing",
    3: "Charge",
    4: "Discharge",
    5: "Fault",
    6: "Preserve Charge",
    7: "Idle",
    10: "Power Saving",
}

EXPORT_CONTROL_MODE_MAP = {
    0: "Disabled",
    1: "Direct Export Limitation",
    2: "Indirect Export Limitation",
    3: "Production Limitation"
}

REACTIVE_POWER_CONFIG_MAP = {
    0: "Fixed CosPhi",
    1: "Fixed Q",
    2: "CosPhi(P)",
    3: "Q(U) + Q(P)",
    4: "RRCR Mode"
}


class SunSpecValueType(EnumModel):
    INT16 = "int16", int, 0x8000, "h"
    UINT16 = "uint16", int, 0xFFFF, "H"
    INT32 = "int32", int, 0x80000000, "i"
    UINT32 = "uint32", int, 0xFFFFFFFF, "I"
    UINT64 = "uint64", int, 0xFFFFFFFFFFFFFFFF, "q"
    FLOAT32 = "float32", float, 0x7FC00000, "f"
    STRING = "string", str, "", "s"

    def __init__(
        self,
        identifier: str,
        typed: Type[SunSpecRawData],
        not_implemented_value: SunSpecRawData,
        data_type: str,
    ):
        # pylint: disable=super-init-not-called
        self._identifier = identifier
        self._typed = typed
        self._not_implemented_value = not_implemented_value
        self._data_type = self.__modbus_data_type(data_type)

    @staticmethod
    def __modbus_data_type(data_type: str) -> ModbusTcpClient.DATATYPE:
        for datatype in ModbusTcpClient.DATATYPE:
            if datatype.value[0] == data_type:
                return datatype

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def typed(self) -> Type[SunSpecRawData]:
        return self._typed

    @property
    def length(self) -> int:
        return self._data_type.value[1]

    @property
    def not_implemented_value(self) -> SunSpecRawData:
        return self._not_implemented_value

    @property
    def data_type(self) -> ModbusTcpClient.DATATYPE:
        return self._data_type



#!/usr/bin/env python3
from pathlib import Path

import tomli

pyproject = Path("pyproject.toml")
with pyproject.open("rb") as f:
    data = tomli.load(f)

project = data.get("project", {})
dependencies = project.get("dependencies", [])
optional = project.get("optional-dependencies", {})

# requirements.txt: alle deps + forecast
requirements_all = sorted(set(dependencies + optional.get("forecast", [])))

# requirements-armv7.txt: nur deps
requirements_arm = sorted(set(dependencies))

Path("requirements.txt").write_text("\n".join(requirements_all) + "\n")
Path("requirements-armv7.txt").write_text("\n".join(requirements_arm) + "\n")

print("✅ requirements.txt und requirements-armv7.txt wurden generiert.")
